# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from abc import ABC, abstractmethod

import awkward as ak
from awkward._kernels import KernelError
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import ArrayLike, NumpyLike, NumpyMetadata
from awkward._singleton import Singleton
from awkward._typing import Callable, Protocol, Tuple, TypeAlias, TypeVar, Unpack

np = NumpyMetadata.instance()
numpy = Numpy.instance()


T = TypeVar("T", covariant=True)
KernelKeyType: TypeAlias = Tuple[str, Unpack[Tuple[np.dtype, ...]]]
KernelType: TypeAlias = "Callable[..., KernelError | None]"


class UfuncLike(Protocol):
    def __call__(self, *args: ArrayLike, **kwargs) -> ArrayLike:
        ...


class Backend(Singleton, ABC):
    name: str

    @property
    @abstractmethod
    def nplike(self) -> NumpyLike:
        raise NotImplementedError

    @property
    @abstractmethod
    def index_nplike(self) -> NumpyLike:
        raise NotImplementedError

    def __getitem__(self, key: KernelKeyType) -> KernelType:
        raise NotImplementedError

    def prepare_reducer(self, reducer: ak._reducers.Reducer) -> ak._reducers.Reducer:
        return reducer

    def prepare_ufunc(self, ufunc: UfuncLike) -> UfuncLike:
        return ufunc

    def format_kernel_error(
        self,
        error: KernelError,
    ):
        if error.filename is None:
            filename = ""
        else:
            filename = " (in compiled code: " + error.filename.decode(
                errors="surrogateescape"
            ).lstrip("\n").lstrip("(")

        message = error.str.decode(errors="surrogateescape")

        if error.attempt != ak._util.kSliceNone:
            message += f" while attempting to get index {error.attempt}"

        message += filename
        return message

    def maybe_kernel_error(self, error: KernelError | None):
        if error is None or error.str is None:
            return
        else:
            raise ValueError(self.format_kernel_error(error))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test():
    array1 = ak.Array(["one", "two", "one", "one"])
    array2 = ak.operations.ak_to_categorical.to_categorical(array1)
    assert array1.type != array2.type
    assert array2.type == ak.types.ArrayType(
        ak.types.ListType(
            ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
            parameters={"__array__": "string", "__categorical__": True},
        ),
        4,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("moment",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def moment(x, n, weight=None, axis=None, *, keepdims=False, mask_identity=False):
    """
    Args:
        x: The data on which to compute the moment (anything #ak.to_layout recognizes).
        n (int): The choice of moment: `0` is a sum of weights, `1` is
            #ak.mean, `2` is #ak.var without subtracting the mean, etc.
        weight: Data that can be broadcasted to `x` to give each value a
            weight. Weighting values equally is the same as no weights;
            weighting some values higher increases the significance of those
            values. Weights can be zero or negative.
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this function decreases the number of
            dimensions by 1; if True, the output values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, the application of this function on
            empty lists results in None (an option type); otherwise, the
            calculation is followed through with the reducers' identities,
            usually resulting in floating-point `nan`.

    Computes the `n`th moment in each group of elements from `x` (many
    types supported, including all Awkward Arrays and Records). The grouping
    is performed the same way as for reducers, though this operation is not a
    reducer and has no identity.

    This function has no NumPy equivalent.

    Passing all arguments to the reducers, the moment is calculated as

        ak.sum((x*weight)**n) / ak.sum(weight)

    The `n=2` moment differs from #ak.var in that #ak.var also subtracts the
    mean (the `n=1` moment).

    See #ak.sum for a complete description of handling nested lists and
    missing values (None) in reducers, and #ak.mean for an example with another
    non-reducer.
    """
    # Dispatch
    yield x, weight

    # Implementation
    return _impl(x, n, weight, axis, keepdims, mask_identity)


def _impl(x, n, weight, axis, keepdims, mask_identity):
    axis = regularize_axis(axis)
    behavior = behavior_of(x, weight)
    x = ak.highlevel.Array(
        ak.operations.to_layout(x, allow_record=False, allow_other=False),
        behavior=behavior,
    )
    if weight is not None:
        weight = ak.highlevel.Array(
            ak.operations.to_layout(weight, allow_record=False, allow_other=False),
            behavior=behavior,
        )

    with np.errstate(invalid="ignore", divide="ignore"):
        if weight is None:
            sumw = ak.operations.ak_count._impl(
                x,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=behavior,
            )
            sumwxn = ak.operations.ak_sum._impl(
                x**n,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=behavior,
            )
        else:
            sumw = ak.operations.ak_sum._impl(
                x * 0 + weight,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=behavior,
            )
            sumwxn = ak.operations.ak_sum._impl(
                (x * weight) ** n,
                axis,
                keepdims,
                mask_identity,
                highlevel=True,
                behavior=behavior,
            )
        return sumwxn / sumw



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak

ROOT = pytest.importorskip("ROOT")

ROOT.ROOT.EnableImplicitMT(1)

compiler = ROOT.gInterpreter.Declare


def test_data_frame_integers():
    ak_array_in = ak.Array([1, 2, 3, 4, 5])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x") == "int64_t"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_double():
    ak_array_in = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x") == "double"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_char():
    ak_array_in = ak.Array(["a", "b", "c", "d", "e"])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x") == "std::string"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_complex():
    ak_array_in = ak.Array([1.1 + 0.1j, 2.2 + 0.2j, 3.3 + 0.3j, 4.4 + 0.4j, 5.5 + 0.5j])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x") == "std::complex<double>"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_listoffset_integers():
    ak_array_in = ak.Array([[1], [2, 3, 4], [5]])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<int64_t>"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_listoffset_listoffset_double():
    ak_array_in = ak.Array(
        [
            [[1.1, 2.2, 3.3]],
            [[4.4, 5.5]],
            [[6.6], [7.7, 8.8, 9.9]],
        ]
    )

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    # awkward::ListArray_ type
    # assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<double>"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_vec_of_vec():
    array = ak.Array(
        [
            [
                {"x": 1.1, "y": [1]},
                {"x": None, "y": [1, 2]},
                {"x": 3.3, "y": [1, 2, 3]},
            ],
            [],
            [{"x": None, "y": [1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}],
        ]
    )
    # ] * 10000)

    rdf2 = ak.to_rdataframe({"array": array})
    # We create a matrix RxC here
    # Note when dimensions R and C are large, the following code suffers
    # from potential performance penalties caused by frequent reallocation
    # of memory by the push_back() function. This should be used only when
    # vector dimensions are not known in advance.
    rdf3 = rdf2.Define(
        "output",
        """
    std::vector<std::vector<double>> tmp1;

    for (auto record : array) {
        std::vector<double> tmp2;
        for (auto number : record.y()) {
            tmp2.push_back(number * number);
        }
        tmp1.push_back(tmp2);
    }
    return tmp1;
    """,
    )

    assert rdf3.GetColumnType("output") == "vector<vector<double> >"
    out = ak.from_rdataframe(
        rdf3,
        columns=("output",),
    )
    assert out["output"].to_list() == (array["y"] * array["y"] * 1.0).to_list()

    rdf3 = rdf2.Define(
        "output2",
        """
    std::vector<std::vector<std::vector<double>>> tmp1;

    for (auto record : array) {
        std::vector<std::vector<double>> tmp2;
        // we can check if it's None:
        // if (record.x().has_value())
        // or set it to 1 so that we do not scale:
        double x_number = record.x().value_or(1);
        for (auto number : record.y()) {
            std::vector<double> tmp3;
            for (int64_t i = 0; i < std::rint(x_number); i++) {
                double value = x_number * number;
                tmp3.push_back(value);
            }
            tmp2.push_back(tmp3);
        }
        tmp1.push_back(tmp2);
    }
    return tmp1;
    """,
    )
    assert rdf3.GetColumnType("output2") == "vector<vector<vector<double> > >"
    out = ak.from_rdataframe(
        rdf3,
        columns=("output2",),
    )
    result = ak.Array(
        [
            [
                [[1.1]],  # "x" is 1 - "y" values are unchanged, and each is nesed
                [
                    [1.0],
                    [2.0],
                ],  # "x" is None - "y" values are unchanged, and each is nesed
                [
                    [3.3, 3.3, 3.3],
                    [6.6, 6.6, 6.6],
                    [9.899999999999999, 9.899999999999999, 9.899999999999999],
                ],  # "x" is 3.3 - "y" values are scaled by 3.3 and each is nesed 3 times
            ],
            [],
            [
                [
                    [1.0],
                    [2.0],
                    [3.0],
                    [4.0],
                ],  # "x" is None - "y" values are unchanged, and each is nesed
                [
                    [5.5, 5.5, 5.5, 5.5, 5.5, 5.5],
                    [11.0, 11.0, 11.0, 11.0, 11.0, 11.0],
                    [16.5, 16.5, 16.5, 16.5, 16.5, 16.5],
                    [22.0, 22.0, 22.0, 22.0, 22.0, 22.0],
                    [27.5, 27.5, 27.5, 27.5, 27.5, 27.5],
                ],  # "x" is 5.5 - "y" values are scaled by 5.5 and each is nesed 5 times
            ],
        ]
    )
    assert out["output2"].to_list() == result.to_list()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()
numpy = Numpy.instance()


class HashableDict:
    def __init__(self, obj):
        self.keys = tuple(sorted(obj))
        self.values = tuple(as_hashable(obj[k]) for k in self.keys)
        self.hash = hash((HashableDict, *self.keys), self.values)

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (
            isinstance(other, HashableDict)
            and self.keys == other.keys
            and self.values == other.values
        )


class HashableList:
    def __init__(self, obj):
        self.values = tuple(obj)
        self.hash = hash((HashableList, *self.values))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return isinstance(other, HashableList) and self.values == other.values


def as_hashable(obj):
    if isinstance(obj, dict):
        return HashableDict(obj)
    elif isinstance(obj, tuple):
        return tuple(as_hashable(x) for x in obj)
    elif isinstance(obj, list):
        return HashableList(obj)
    else:
        return obj



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.numpy import Numpy

numpy = Numpy.instance()


class MyBehavior(ak.Array):
    ...


def test():
    behavior = {"MyBehavior": MyBehavior}
    array = ak.with_parameter([[1, 2, 3]], "__list__", "MyBehavior", behavior=behavior)
    assert isinstance(array, MyBehavior)

    shallow_copy = ak.Array(array)
    assert isinstance(shallow_copy, MyBehavior)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    aa = ak.contents.NumpyArray(np.frombuffer(b"hellothere", "u1"))
    b = aa.to_backend_array(allow_missing=False)
    assert b.tolist() == [104, 101, 108, 108, 111, 116, 104, 101, 114, 101]
    assert b.dtype == np.dtype(np.uint8)

    c = ak.contents.NumpyArray(np.array([0, 1577836800], dtype="datetime64[s]"))
    assert [
        d.isoformat() for d in c.to_backend_array(allow_missing=False).tolist()
    ] == [
        "1970-01-01T00:00:00",
        "2020-01-01T00:00:00",
    ]

    recordarray = ak.contents.RecordArray(
        [ak.contents.NumpyArray(np.array([1, 2, 3, 4, 5], dtype=np.int64))],
        fields=["one"],
    )
    assert recordarray.to_backend_array(allow_missing=False).tolist() == [
        (1,),
        (2,),
        (3,),
        (4,),
        (5,),
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("cartesian",)
from collections.abc import Mapping

import awkward as ak
from awkward._backends.dispatch import backend_of
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._errors import AxisError
from awkward._layout import maybe_posaxis, wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()
cpu = NumpyBackend.instance()


@high_level_function
def cartesian(
    arrays,
    axis=1,
    *,
    nested=None,
    parameters=None,
    with_name=None,
    highlevel=True,
    behavior=None,
):
    """
    Args:
        arrays (dict or iterable of arrays): Each value in this dict or iterable
            can be any array-like data that #ak.to_layout recognizes.
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        nested (None, True, False, or iterable of str or int): If None or
            False, all combinations of elements from the `arrays` are
            produced at the same level of nesting; if True, they are grouped
            in nested lists by combinations that share a common item from
            each of the `arrays`; if an iterable of str or int, group common
            items for a chosen set of keys from the `array` dict or integer
            slots of the `array` iterable.
        parameters (None or dict): Parameters for the new
            #ak.contents.RecordArray node that is created by this operation.
        with_name (None or str): Assigns a `"__record__"` name to the new
            #ak.contents.RecordArray node that is created by this operation
            (overriding `parameters`, if necessary).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Computes a Cartesian product (i.e. cross product) of data from a set of
    `arrays`. This operation creates records (if `arrays` is a dict) or tuples
    (if `arrays` is another kind of iterable) that hold the combinations
    of elements, and it can introduce new levels of nesting.

    As a simple example with `axis=0`, the Cartesian product of

        >>> one = ak.Array([1, 2, 3])
        >>> two = ak.Array(["a", "b"])

    is

        >>> ak.cartesian([one, two], axis=0).show()
        [(1, 'a'),
         (1, 'b'),
         (2, 'a'),
         (2, 'b'),
         (3, 'a'),
         (3, 'b')]

    With nesting, a new level of nested lists is created to group combinations
    that share the same element from `one` into the same list.

        >>> ak.cartesian([one, two], axis=0, nested=True).show()
        [[(1, 'a'), (1, 'b')],
         [(2, 'a'), (2, 'b')],
         [(3, 'a'), (3, 'b')]]

    The primary purpose of this function, however, is to compute a different
    Cartesian product for each element of an array: in other words, `axis=1`.
    The following arrays each have four elements.

        >>> one = ak.Array([[1, 2, 3], [], [4, 5], [6]])
        >>> two = ak.Array([["a", "b"], ["c"], ["d"], ["e", "f"]])

    The default `axis=1` produces 6 pairs from the Cartesian product of
    `[1, 2, 3]` and `["a", "b"]`, 0 pairs from `[]` and `["c"]`, 1 pair from
    `[4, 5]` and `["d"]`, and 1 pair from `[6]` and `["e", "f"]`.

        >>> ak.cartesian([one, two]).show()
        [[(1, 'a'), (1, 'b'), (2, 'a'), (2, 'b'), (3, 'a'), (3, 'b')],
         [],
         [(4, 'd'), (5, 'd')],
         [(6, 'e'), (6, 'f')]]

    The nesting depth is the same as the original arrays; with `nested=True`,
    the nesting depth is increased by 1 and tuples are grouped by their
    first element.

        >>> ak.cartesian([one, two], nested=True).show()
        [[[(1, 'a'), (1, 'b')], [(2, 'a'), (2, ...)], [(3, 'a'), (3, 'b')]],
         [],
         [[(4, 'd')], [(5, 'd')]],
         [[(6, 'e'), (6, 'f')]]]

    These tuples are #ak.contents.RecordArray nodes with unnamed fields. To
    name the fields, we can pass `one` and `two` in a dict, rather than a list.

        >>> ak.cartesian({"x": one, "y": two}).show()
        [[{x: 1, y: 'a'}, {x: 1, y: 'b'}, {...}, ..., {x: 3, y: 'a'}, {x: 3, y: 'b'}],
         [],
         [{x: 4, y: 'd'}, {x: 5, y: 'd'}],
         [{x: 6, y: 'e'}, {x: 6, y: 'f'}]]

    With more than two elements in the Cartesian product, `nested` can specify
    which are grouped and which are not. For example,

        >>> one = ak.Array([1, 2, 3, 4])
        >>> two = ak.Array([1.1, 2.2, 3.3])
        >>> three = ak.Array(["a", "b"])

    can be left entirely ungrouped:

        >>> ak.cartesian([one, two, three], axis=0).show()
        [(1, 1.1, 'a'),
         (1, 1.1, 'b'),
         (1, 2.2, 'a'),
         (1, 2.2, 'b'),
         (1, 3.3, 'a'),
         (1, 3.3, 'b'),
         (2, 1.1, 'a'),
         (2, 1.1, 'b'),
         (2, 2.2, 'a'),
         (2, 2.2, 'b'),
         ...,
         (3, 2.2, 'b'),
         (3, 3.3, 'a'),
         (3, 3.3, 'b'),
         (4, 1.1, 'a'),
         (4, 1.1, 'b'),
         (4, 2.2, 'a'),
         (4, 2.2, 'b'),
         (4, 3.3, 'a'),
         (4, 3.3, 'b')]

    can be grouped by `one` (adding 1 more dimension):

        >>> ak.cartesian([one, two, three], axis=0, nested=[0]).show()
        [[(1, 1.1, 'a'), (1, 1.1, 'b'), (1, 2.2, 'a')],
         [(1, 2.2, 'b'), (1, 3.3, 'a'), (1, 3.3, 'b')],
         [(2, 1.1, 'a'), (2, 1.1, 'b'), (2, 2.2, 'a')],
         [(2, 2.2, 'b'), (2, 3.3, 'a'), (2, 3.3, 'b')],
         [(3, 1.1, 'a'), (3, 1.1, 'b'), (3, 2.2, 'a')],
         [(3, 2.2, 'b'), (3, 3.3, 'a'), (3, 3.3, 'b')],
         [(4, 1.1, 'a'), (4, 1.1, 'b'), (4, 2.2, 'a')],
         [(4, 2.2, 'b'), (4, 3.3, 'a'), (4, 3.3, 'b')]]

    can be grouped by `one` and `two` (adding 2 more dimensions):

        >>> ak.cartesian([one, two, three], axis=0, nested=[0, 1]).show()
        [[[(1, 1.1, 'a'), (1, 1.1, 'b')], [...], [(1, 3.3, 'a'), (1, 3.3, ...)]],
         [[(2, 1.1, 'a'), (2, 1.1, 'b')], [...], [(2, 3.3, 'a'), (2, 3.3, ...)]],
         [[(3, 1.1, 'a'), (3, 1.1, 'b')], [...], [(3, 3.3, 'a'), (3, 3.3, ...)]],
         [[(4, 1.1, 'a'), (4, 1.1, 'b')], [...], [(4, 3.3, 'a'), (4, 3.3, ...)]]]

    or grouped by unique `one`-`two` pairs (adding 1 more dimension):

        >>> ak.cartesian([one, two, three], axis=0, nested=[1]).show()
        [[(1, 1.1, 'a'), (1, 1.1, 'b')],
         [(1, 2.2, 'a'), (1, 2.2, 'b')],
         [(1, 3.3, 'a'), (1, 3.3, 'b')],
         [(2, 1.1, 'a'), (2, 1.1, 'b')],
         [(2, 2.2, 'a'), (2, 2.2, 'b')],
         [(2, 3.3, 'a'), (2, 3.3, 'b')],
         [(3, 1.1, 'a'), (3, 1.1, 'b')],
         [(3, 2.2, 'a'), (3, 2.2, 'b')],
         [(3, 3.3, 'a'), (3, 3.3, 'b')],
         [(4, 1.1, 'a'), (4, 1.1, 'b')],
         [(4, 2.2, 'a'), (4, 2.2, 'b')],
         [(4, 3.3, 'a'), (4, 3.3, 'b')]]

    The order of the output is fixed: it is always lexicographical in the
    order that the `arrays` are written.

    To emulate an SQL or Pandas "group by" operation, put the keys that you
    wish to group by *first* and use `nested=[0]` or `nested=[n]` to group by
    unique n-tuples. If necessary, record keys can later be reordered with a
    list of strings in #ak.Array.__getitem__.

    To get list index positions in the tuples/records, rather than data from
    the original `arrays`, use #ak.argcartesian instead of #ak.cartesian. The
    #ak.argcartesian form can be particularly useful as nested indexing in
    #ak.Array.__getitem__.
    """
    # Dispatch
    if isinstance(arrays, Mapping):
        yield arrays.values()
    else:
        yield arrays

    # Implementation
    return _impl(arrays, axis, nested, parameters, with_name, highlevel, behavior)


def _impl(arrays, axis, nested, parameters, with_name, highlevel, behavior):
    axis = regularize_axis(axis)
    if isinstance(arrays, dict):
        backend = backend_of(*arrays.values(), default=cpu)
        behavior = behavior_of(*arrays.values(), behavior=behavior)
        array_layouts = {
            name: ak.operations.to_layout(
                layout, allow_record=False, allow_other=False
            ).to_backend(backend)
            for name, layout in arrays.items()
        }
        layouts = list(array_layouts.values())
        fields = list(array_layouts.keys())

    else:
        arrays = list(arrays)
        backend = backend_of(*arrays, default=cpu)
        behavior = behavior_of(*arrays, behavior=behavior)
        array_layouts = [
            ak.operations.to_layout(
                layout, allow_record=False, allow_other=False
            ).to_backend(backend)
            for layout in arrays
        ]
        layouts = array_layouts
        fields = None

    if with_name is not None:
        if parameters is None:
            parameters = {}
        else:
            parameters = dict(parameters)
        parameters["__record__"] = with_name

    posaxis = maybe_posaxis(layouts[0], axis, 1)
    # Validate `posaxis`
    if posaxis is None or posaxis < 0:
        raise AxisError("negative axis depth is ambiguous")
    # Ensure other layouts have same positive value for axis
    for layout in layouts[1:]:
        if maybe_posaxis(layout, axis, 1) != posaxis:
            raise AxisError(
                "arrays to cartesian-product do not have the same depth for negative axis"
            )
    depths = [obj.purelist_depth for obj in layouts]
    if posaxis >= max(depths):
        raise AxisError(
            f"axis={axis} exceeds the max depth of the given arrays (which is {max(depths)})"
        )

    # Validate `nested`
    if nested is None or nested is False:
        nested = []
    elif nested is True:
        if fields is not None:
            nested = list(fields)[:-1]
        else:
            nested = list(range(len(layouts))[:-1])
    else:
        if isinstance(array_layouts, dict):
            if any(not (isinstance(x, str) and x in array_layouts) for x in nested):
                raise ValueError(
                    "the 'nested' parameter of cartesian must be dict keys "
                    "for a dict of arrays"
                )
            if len(nested) >= len(array_layouts):
                raise ValueError(
                    "the `nested` parameter of cartesian must contain "
                    "fewer items than there are arrays"
                )
        else:
            if any(
                not (isinstance(x, int) and 0 <= x < len(array_layouts) - 1)
                for x in nested
            ):
                raise ValueError(
                    "the 'nested' parameter of cartesian must be integers in "
                    "[0, len(arrays) - 1) for an iterable of arrays"
                )

    if posaxis == 0:
        if fields is not None:
            nested = [i for i, name in enumerate(fields) if name in nested]

        indexes = [
            ak.index.Index64(backend.index_nplike.reshape(x, (-1,)))
            for x in backend.index_nplike.meshgrid(
                *[
                    backend.index_nplike.arange(x.length, dtype=np.int64)
                    for x in layouts
                ],
                indexing="ij",
            )
        ]
        outs = [
            ak.contents.IndexedArray.simplified(x, y) for x, y in zip(indexes, layouts)
        ]

        result = ak.contents.RecordArray(outs, fields, parameters=parameters)
        for i in range(len(array_layouts))[::-1]:
            if i in nested:
                result = ak.contents.RegularArray(result, layouts[i + 1].length, 0)

    else:

        def add_outer_dimensions(
            layout: ak.contents.Content, n: int
        ) -> ak.contents.Content:
            if n == 0:
                return layout
            else:
                return ak.contents.RegularArray(
                    add_outer_dimensions(layout, n - 1), 1, 0
                )

        def apply_pad_inner_list(layout, depth, lateral_context, **kwargs):
            """
            Add new dimensions (given by lateral_context["n"]) above innermost list
            """
            n = lateral_context["n"]
            # We want to be above at least one dimension (list)
            if depth == 2:
                return add_outer_dimensions(layout, n)
            else:
                return None

        def apply_pad_inner_list_at_axis(layout, depth, lateral_context, **kwargs):
            """
            Each array in arrays contributes to one of these new dimensions.
            To make the cartesian product of the given arrays broadcastable,
            each array is padded by (n, m) new length-1 regular dimensions
            (above, below) the target depth. The values of (n, m) are given by
            the position of the array; the first array is the outermost axis.
            """
            i = lateral_context["i"]
            if depth == posaxis:
                n_inside = len(array_layouts) - i - 1
                n_outside = i
                if (
                    layout.parameter("__array__") == "string"
                    or layout.parameter("__array__") == "bytestring"
                ):
                    raise ValueError(
                        "ak.cartesian does not compute combinations of the "
                        "characters of a string; please split it into lists"
                    )
                nextlayout = ak._do.recursively_apply(
                    layout,
                    apply_pad_inner_list,
                    behavior,
                    lateral_context={"n": n_inside},
                )
                return add_outer_dimensions(nextlayout, n_outside)
            else:
                return None

        # New _interior_ axes are added to the result layout, but
        # unless explicitly named, these axes should be flattened.
        axes_to_flatten = [
            posaxis + i + 1
            for i, _ in enumerate(array_layouts)
            if i < len(array_layouts) - 1 and i not in nested
        ]
        # This list *must* be sorted in reverse order
        axes_to_flatten.reverse()

        new_layouts = [
            ak._do.recursively_apply(
                layout,
                apply_pad_inner_list_at_axis,
                behavior,
                lateral_context={"i": i},
            )
            for i, layout in enumerate(layouts)
        ]

        def apply_build_record(inputs, depth, **kwargs):
            if depth == posaxis + len(array_layouts):
                return (ak.contents.RecordArray(inputs, fields, parameters=parameters),)

            else:
                return None

        out = ak._broadcasting.broadcast_and_apply(
            new_layouts, apply_build_record, behavior, right_broadcast=False
        )
        assert isinstance(out, tuple) and len(out) == 1
        result = out[0]

        # Remove surplus dimensions, iterating from smallest to greatest
        for axis_to_flatten in axes_to_flatten:
            result = ak.operations.flatten(
                result, axis=axis_to_flatten, highlevel=False, behavior=behavior
            )

    return wrap_layout(result, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("drop_none",)
import awkward as ak
from awkward._dispatch import high_level_function
from awkward._errors import AxisError
from awkward._layout import maybe_posaxis, wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def drop_none(array, axis=None, highlevel=True, behavior=None):
    """
    Args:
        array: Data in which to remove Nones.
        axis (None or int): If None, the operation drops Nones at all levels of
            nesting, returning an array of the same dimension, but without Nones.
            Otherwise, it drops Nones at a specified depth.
            The outermost dimension is `0`, followed by `1`, etc.,
            and negative values count backward from the innermost: `-1` is the
            innermost dimension, `-2` is the next level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Removes missing values (None) from a given array.

    For example, in the following `array`,

        >>> array = ak.Array([[[0]], [[None]], [[1], None], [[2, None]]])

    The None value will be removed, resulting in

        >>> ak.drop_none(array)
        <Array [[[0]], [[]], [[1]], [[2]]] type='4 * var * var * int64'>

    The default axis is None, however an axis can be specified:

        >>> ak.drop_none(array, axis=1)
        <Array [[[0]], [[None]], [[1]], [[2, None]]] type='4 * var * var * ?int64'>

    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior)


def _impl(array, axis, highlevel, behavior):
    axis = regularize_axis(axis)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)

    def drop_nones(layout, **kwargs):
        if layout.is_list:
            # only drop nones at list level in the recursion; this way ListArray -> ListOffsetArray with unprojected optiontype -> avoid offset mismatch
            return layout.drop_none()
        else:
            return layout

    if axis is None:
        # if the outer layout is_option, drop_nones without affecting offsets
        if layout.is_option:
            layout = layout.drop_none()

        def action(layout, continuation, **kwargs):
            return drop_nones(continuation())

    else:
        max_axis = layout.branch_depth[1] - 1
        if axis > max_axis:
            raise AxisError(f"axis={axis} exceeds the depth ({max_axis}) of this array")

        def recompute_offsets(layout, depth, **kwargs):
            posaxis = maybe_posaxis(layout, axis, depth)
            if (
                posaxis == 0
                and posaxis == depth - 1
                or posaxis == depth
                and layout.is_list
            ):
                none_indexes = options["none_indexes"].pop(0)
                out = layout._rebuild_without_nones(none_indexes, layout.content)
                return out

        def action(layout, depth, **kwargs):
            if layout.is_record:
                posaxises = {maybe_posaxis(x, axis, depth) for x in layout.contents}
                if len(posaxises) > 1 and any(x < depth for x in posaxises):
                    raise AxisError(
                        f"axis={axis} implies different levels in records that might require part of a record to be dropped, which is impossible"
                    )
            posaxis = maybe_posaxis(layout, axis, depth)
            if posaxis == 0:
                if not layout.is_option:
                    return layout
                else:
                    return layout.drop_none()
            if posaxis == depth - 1 and layout.is_option:
                _, _, none_indexes = layout._nextcarry_outindex()
                options["none_indexes"].append(none_indexes)
                return layout.drop_none()
            if posaxis == depth - 1 and layout.is_list and layout.content.is_option:
                return layout.drop_none()

    options = {"none_indexes": []}
    out = ak._do.recursively_apply(layout, action, behavior, options)

    if len(options["none_indexes"]) > 0:
        out = ak._do.recursively_apply(out, recompute_offsets, behavior, options)

    return wrap_layout(out, behavior, highlevel, like=behavior)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
# ruff: noqa: F401

from awkward.operations.ak_all import *
from awkward.operations.ak_almost_equal import *
from awkward.operations.ak_any import *
from awkward.operations.ak_argcartesian import *
from awkward.operations.ak_argcombinations import *
from awkward.operations.ak_argmax import *
from awkward.operations.ak_argmin import *
from awkward.operations.ak_argsort import *
from awkward.operations.ak_backend import *
from awkward.operations.ak_broadcast_arrays import *
from awkward.operations.ak_broadcast_fields import *
from awkward.operations.ak_cartesian import *
from awkward.operations.ak_categories import *
from awkward.operations.ak_combinations import *
from awkward.operations.ak_concatenate import *
from awkward.operations.ak_copy import *
from awkward.operations.ak_corr import *
from awkward.operations.ak_count import *
from awkward.operations.ak_count_nonzero import *
from awkward.operations.ak_covar import *
from awkward.operations.ak_drop_none import *
from awkward.operations.ak_enforce_type import *
from awkward.operations.ak_fields import *
from awkward.operations.ak_fill_none import *
from awkward.operations.ak_firsts import *
from awkward.operations.ak_flatten import *
from awkward.operations.ak_from_arrow import *
from awkward.operations.ak_from_arrow_schema import *
from awkward.operations.ak_from_avro_file import *
from awkward.operations.ak_from_buffers import *
from awkward.operations.ak_from_categorical import *
from awkward.operations.ak_from_cupy import *
from awkward.operations.ak_from_iter import *
from awkward.operations.ak_from_jax import *
from awkward.operations.ak_from_json import *
from awkward.operations.ak_from_numpy import *
from awkward.operations.ak_from_parquet import *
from awkward.operations.ak_from_rdataframe import *
from awkward.operations.ak_from_regular import *
from awkward.operations.ak_full_like import *
from awkward.operations.ak_is_categorical import *
from awkward.operations.ak_is_none import *
from awkward.operations.ak_is_tuple import *
from awkward.operations.ak_is_valid import *
from awkward.operations.ak_isclose import *
from awkward.operations.ak_linear_fit import *
from awkward.operations.ak_local_index import *
from awkward.operations.ak_mask import *
from awkward.operations.ak_max import *
from awkward.operations.ak_mean import *
from awkward.operations.ak_merge_option_of_records import *
from awkward.operations.ak_merge_union_of_records import *
from awkward.operations.ak_metadata_from_parquet import *
from awkward.operations.ak_min import *
from awkward.operations.ak_moment import *
from awkward.operations.ak_nan_to_none import *
from awkward.operations.ak_nan_to_num import *
from awkward.operations.ak_num import *
from awkward.operations.ak_ones_like import *
from awkward.operations.ak_pad_none import *
from awkward.operations.ak_parameters import *
from awkward.operations.ak_prod import *
from awkward.operations.ak_ptp import *
from awkward.operations.ak_ravel import *
from awkward.operations.ak_run_lengths import *
from awkward.operations.ak_singletons import *
from awkward.operations.ak_softmax import *
from awkward.operations.ak_sort import *
from awkward.operations.ak_std import *
from awkward.operations.ak_strings_astype import *
from awkward.operations.ak_sum import *
from awkward.operations.ak_to_arrow import *
from awkward.operations.ak_to_arrow_table import *
from awkward.operations.ak_to_backend import *
from awkward.operations.ak_to_buffers import *
from awkward.operations.ak_to_categorical import *
from awkward.operations.ak_to_cupy import *
from awkward.operations.ak_to_dataframe import *
from awkward.operations.ak_to_jax import *
from awkward.operations.ak_to_json import *
from awkward.operations.ak_to_layout import *
from awkward.operations.ak_to_list import *
from awkward.operations.ak_to_numpy import *
from awkward.operations.ak_to_packed import *
from awkward.operations.ak_to_parquet import *
from awkward.operations.ak_to_rdataframe import *
from awkward.operations.ak_to_regular import *
from awkward.operations.ak_transform import *
from awkward.operations.ak_type import *
from awkward.operations.ak_unflatten import *
from awkward.operations.ak_unzip import *
from awkward.operations.ak_validity_error import *
from awkward.operations.ak_values_astype import *
from awkward.operations.ak_var import *
from awkward.operations.ak_where import *
from awkward.operations.ak_with_field import *
from awkward.operations.ak_with_name import *
from awkward.operations.ak_with_parameter import *
from awkward.operations.ak_without_field import *
from awkward.operations.ak_without_parameters import *
from awkward.operations.ak_zeros_like import *
from awkward.operations.ak_zip import *



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_boolean():
    result = ak.operations.from_json(
        " [ true ,false, true, true, false]    ",
        schema={"type": "array", "items": {"type": "boolean"}},
    )
    assert result.to_list() == [True, False, True, True, False]
    assert str(result.type) == "5 * bool"

    result = ak.operations.from_json(
        " [ true ,false, true, true, false]    " * 2,
        schema={"type": "array", "items": {"type": "boolean"}},
        line_delimited=True,
    )
    assert result.to_list() == [True, False, True, True, False] * 2
    assert str(result.type) == "10 * bool"

    result = ak.operations.from_json(
        "",
        schema={"type": "array", "items": {"type": "boolean"}},
        line_delimited=True,
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * bool"

    result = ak.operations.from_json(
        "[]",
        schema={"type": "array", "items": {"type": "boolean"}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * bool"


def test_integer():
    with pytest.raises(ValueError):
        ak.operations.from_json(
            " [ 1 ,2 ,3.0, 4, 5]  \n  ",
            schema={"type": "array", "items": {"type": "integer"}},
        )

    result = ak.operations.from_json(
        " [ 1 ,2 ,3, 4, 5]  \n  ",
        schema={"type": "array", "items": {"type": "integer"}},
    )
    assert result.to_list() == [1, 2, 3, 4, 5]
    assert str(result.type) == "5 * int64"

    result = ak.operations.from_json(
        " [ 1 ,2 ,3, 4, 5]  \n  " * 2,
        schema={"type": "array", "items": {"type": "integer"}},
        line_delimited=True,
    )
    assert result.to_list() == [1, 2, 3, 4, 5] * 2
    assert str(result.type) == "10 * int64"

    result = ak.operations.from_json(
        "[ ]",
        schema={"type": "array", "items": {"type": "integer"}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * int64"


def test_number():
    result = ak.operations.from_json(
        " [ 1 ,2,3.14, 4, 5]",
        schema={"type": "array", "items": {"type": "number"}},
    )
    assert result.to_list() == [1, 2, 3.14, 4, 5]
    assert str(result.type) == "5 * float64"

    result = ak.operations.from_json(
        " [ 1 ,2,3.14, 4, 5]" * 2,
        schema={"type": "array", "items": {"type": "number"}},
        line_delimited=True,
    )
    assert result.to_list() == [1, 2, 3.14, 4, 5] * 2
    assert str(result.type) == "10 * float64"

    result = ak.operations.from_json(
        " [ ]",
        schema={"type": "array", "items": {"type": "number"}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * float64"


def test_option_boolean():
    result = ak.operations.from_json(
        " [ true ,false ,null , true, false]",
        schema={"type": "array", "items": {"type": ["boolean", "null"]}},
    )
    assert result.to_list() == [True, False, None, True, False]
    assert str(result.type) == "5 * ?bool"

    result = ak.operations.from_json(
        " [ true ,false ,null , true, false]" * 2,
        schema={"type": "array", "items": {"type": ["boolean", "null"]}},
        line_delimited=True,
    )
    assert result.to_list() == [True, False, None, True, False] * 2
    assert str(result.type) == "10 * ?bool"

    result = ak.operations.from_json(
        " [ ]",
        schema={"type": "array", "items": {"type": ["boolean", "null"]}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * ?bool"


def test_option_integer():
    result = ak.operations.from_json(
        " [ 1 ,2,null,4, 5]",
        schema={"type": "array", "items": {"type": ["null", "integer"]}},
    )
    assert result.to_list() == [1, 2, None, 4, 5]
    assert str(result.type) == "5 * ?int64"

    result = ak.operations.from_json(
        " [ 1 ,2,null,4, 5]" * 2,
        schema={"type": "array", "items": {"type": ["null", "integer"]}},
        line_delimited=True,
    )
    assert result.to_list() == [1, 2, None, 4, 5] * 2
    assert str(result.type) == "10 * ?int64"

    result = ak.operations.from_json(
        " [ ]",
        schema={"type": "array", "items": {"type": ["null", "integer"]}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * ?int64"


def test_string():
    result = ak.operations.from_json(
        r' [ "" ,"two","three \u2192 3", "\"four\"", "fi\nve"]',
        schema={"type": "array", "items": {"type": "string"}},
    )
    assert result.to_list() == ["", "two", "three \u2192 3", '"four"', "fi\nve"]
    assert str(result.type) == "5 * string"

    result = ak.operations.from_json(
        r' [ "" ,"two","three \u2192 3", "\"four\"", "fi\nve"]' * 2,
        schema={"type": "array", "items": {"type": "string"}},
        line_delimited=True,
    )
    assert result.to_list() == ["", "two", "three \u2192 3", '"four"', "fi\nve"] * 2
    assert str(result.type) == "10 * string"

    result = ak.operations.from_json(
        r"[]",
        schema={"type": "array", "items": {"type": "string"}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * string"


def test_option_string():
    result = ak.operations.from_json(
        r' [ "" ,null ,"three \u2192 3", "\"four\"", "fi\nve"]',
        schema={"type": "array", "items": {"type": ["null", "string"]}},
    )
    assert result.to_list() == ["", None, "three \u2192 3", '"four"', "fi\nve"]
    assert str(result.type) == "5 * ?string"

    result = ak.operations.from_json(
        r' [ "" ,null ,"three \u2192 3", "\"four\"", "fi\nve"]' * 2,
        schema={"type": "array", "items": {"type": ["null", "string"]}},
        line_delimited=True,
    )
    assert result.to_list() == ["", None, "three \u2192 3", '"four"', "fi\nve"] * 2
    assert str(result.type) == "10 * ?string"

    result = ak.operations.from_json(
        r"[]",
        schema={"type": "array", "items": {"type": ["null", "string"]}},
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * ?string"


def test_enum_string():
    result = ak.operations.from_json(
        r'["three", "two", "one", "one", "two", "three"]',
        schema={
            "type": "array",
            "items": {"type": "string", "enum": ["one", "two", "three"]},
        },
    )
    assert result.to_list() == ["three", "two", "one", "one", "two", "three"]
    assert isinstance(result.layout, ak.contents.IndexedArray)
    assert result.layout.index.data.tolist() == [2, 1, 0, 0, 1, 2]
    assert str(result.type) == "6 * categorical[type=string]"

    result = ak.operations.from_json(
        r'["three", "two", "one", "one", "two", "three"]' * 2,
        schema={
            "type": "array",
            "items": {"type": "string", "enum": ["one", "two", "three"]},
        },
        line_delimited=True,
    )
    assert result.to_list() == ["three", "two", "one", "one", "two", "three"] * 2
    assert isinstance(result.layout, ak.contents.IndexedArray)
    assert result.layout.index.data.tolist() == [2, 1, 0, 0, 1, 2] * 2
    assert str(result.type) == "12 * categorical[type=string]"

    result = ak.operations.from_json(
        r"[]",
        schema={
            "type": "array",
            "items": {"type": "string", "enum": ["one", "two", "three"]},
        },
    )
    assert result.to_list() == []
    assert isinstance(result.layout, ak.contents.IndexedArray)
    assert result.layout.index.data.tolist() == []
    assert str(result.type) == "0 * categorical[type=string]"


def test_option_enum_string():
    result = ak.operations.from_json(
        r'["three", "two", null, "one", "one", "two", "three"]',
        schema={
            "type": "array",
            "items": {"type": ["null", "string"], "enum": ["one", "two", "three"]},
        },
    )
    assert result.to_list() == ["three", "two", None, "one", "one", "two", "three"]
    assert isinstance(result.layout, ak.contents.IndexedOptionArray)
    assert result.layout.index.data.tolist() == [2, 1, -1, 0, 0, 1, 2]
    assert str(result.type) == "7 * ?categorical[type=string]"

    result = ak.operations.from_json(
        r'["three", "two", null, "one", "one", "two", "three"]' * 2,
        schema={
            "type": "array",
            "items": {"type": ["null", "string"], "enum": ["one", "two", "three"]},
        },
        line_delimited=True,
    )
    assert result.to_list() == ["three", "two", None, "one", "one", "two", "three"] * 2
    assert isinstance(result.layout, ak.contents.IndexedOptionArray)
    assert result.layout.index.data.tolist() == [2, 1, -1, 0, 0, 1, 2] * 2
    assert str(result.type) == "14 * ?categorical[type=string]"

    result = ak.operations.from_json(
        r"[]",
        schema={
            "type": "array",
            "items": {"type": ["null", "string"], "enum": ["one", "two", "three"]},
        },
    )
    assert result.to_list() == []
    assert isinstance(result.layout, ak.contents.IndexedOptionArray)
    assert result.layout.index.data.tolist() == []
    assert str(result.type) == "0 * ?categorical[type=string]"


def test_array_integer():
    result = ak.operations.from_json(
        " [ [ 1 ,2, 3], [], [4, 5]]",
        schema={
            "type": "array",
            "items": {"type": "array", "items": {"type": "integer"}},
        },
    )
    assert result.to_list() == [[1, 2, 3], [], [4, 5]]
    assert str(result.type) == "3 * var * int64"

    result = ak.operations.from_json(
        " [ [ 1 ,2, 3], [], [4, 5]]" * 2,
        schema={
            "type": "array",
            "items": {"type": "array", "items": {"type": "integer"}},
        },
        line_delimited=True,
    )
    assert result.to_list() == [[1, 2, 3], [], [4, 5]] * 2
    assert str(result.type) == "6 * var * int64"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {"type": "array", "items": {"type": "integer"}},
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * var * int64"


def test_regulararray_integer():
    result = ak.operations.from_json(
        "[[1, 2, 3], [4, 5, 6]]",
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
    )
    assert result.to_list() == [[1, 2, 3], [4, 5, 6]]
    assert str(result.type) == "2 * 3 * int64"

    result = ak.operations.from_json(
        "[[1, 2, 3], [4, 5, 6]]" * 2,
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
        line_delimited=True,
    )
    assert result.to_list() == [[1, 2, 3], [4, 5, 6]] * 2
    assert str(result.type) == "4 * 3 * int64"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * 3 * int64"


def test_option_regulararray_integer():
    result = ak.operations.from_json(
        "[[1, 2, 3], null, [4, 5, 6]]",
        schema={
            "type": "array",
            "items": {
                "type": ["array", "null"],
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
    )
    assert result.to_list() == [[1, 2, 3], None, [4, 5, 6]]
    assert str(result.type) == "3 * option[3 * int64]"

    result = ak.operations.from_json(
        "[[1, 2, 3], null, [4, 5, 6]]" * 2,
        schema={
            "type": "array",
            "items": {
                "type": ["array", "null"],
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
        line_delimited=True,
    )
    assert result.to_list() == [[1, 2, 3], None, [4, 5, 6]] * 2
    assert str(result.type) == "6 * option[3 * int64]"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {
                "type": ["array", "null"],
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * option[3 * int64]"


def test_option_array_integer():
    result = ak.operations.from_json(
        " [ [ 1 ,2,3 ],null,[ ], [4, 5]]",
        schema={
            "type": "array",
            "items": {"type": ["null", "array"], "items": {"type": "integer"}},
        },
    )
    assert result.to_list() == [[1, 2, 3], None, [], [4, 5]]
    assert str(result.type) == "4 * option[var * int64]"

    result = ak.operations.from_json(
        " [ [ 1 ,2,3 ],null,[ ], [4, 5]]" * 2,
        schema={
            "type": "array",
            "items": {"type": ["null", "array"], "items": {"type": "integer"}},
        },
        line_delimited=True,
    )
    assert result.to_list() == [[1, 2, 3], None, [], [4, 5]] * 2
    assert str(result.type) == "8 * option[var * int64]"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {"type": ["null", "array"], "items": {"type": "integer"}},
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * option[var * int64]"


def test_option_array_option_integer():
    result = ak.operations.from_json(
        " [ [ 1 ,2,3 ],null,[ ] ,[null, 5]]",
        schema={
            "type": "array",
            "items": {
                "type": ["null", "array"],
                "items": {"type": ["integer", "null"]},
            },
        },
    )
    assert result.to_list() == [[1, 2, 3], None, [], [None, 5]]
    assert str(result.type) == "4 * option[var * ?int64]"

    result = ak.operations.from_json(
        " [ [ 1 ,2,3 ],null,[ ] ,[null, 5]]" * 2,
        schema={
            "type": "array",
            "items": {
                "type": ["null", "array"],
                "items": {"type": ["integer", "null"]},
            },
        },
        line_delimited=True,
    )
    assert result.to_list() == [[1, 2, 3], None, [], [None, 5]] * 2
    assert str(result.type) == "8 * option[var * ?int64]"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {
                "type": ["null", "array"],
                "items": {"type": ["integer", "null"]},
            },
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * option[var * ?int64]"


def test_array_array_integer():
    result = ak.operations.from_json(
        " [ [ [ 1 ,2,3 ] ] ,[ [], [4, 5]], []]",
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "array", "items": {"type": "integer"}},
            },
        },
    )
    assert result.to_list() == [[[1, 2, 3]], [[], [4, 5]], []]
    assert str(result.type) == "3 * var * var * int64"

    result = ak.operations.from_json(
        " [ [ [ 1 ,2,3 ] ] ,[ [], [4, 5]], []]" * 2,
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "array", "items": {"type": "integer"}},
            },
        },
        line_delimited=True,
    )
    assert result.to_list() == [[[1, 2, 3]], [[], [4, 5]], []] * 2
    assert str(result.type) == "6 * var * var * int64"

    result = ak.operations.from_json(
        " [ [  ] ,[ ], []]",
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "array", "items": {"type": "integer"}},
            },
        },
    )
    assert result.to_list() == [[], [], []]
    assert str(result.type) == "3 * var * var * int64"

    result = ak.operations.from_json(
        " [ ]",
        schema={
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "array", "items": {"type": "integer"}},
            },
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * var * var * int64"


def test_record():
    result = ak.operations.from_json(
        ' [ { "x" :1 ,"y":1.1},{"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]',
        schema={
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.to_list() == [
        {"x": 1, "y": 1.1},
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]
    assert str(result.type) == "3 * {x: int64, y: float64}"

    result = ak.operations.from_json(
        ' [ { "x" :1 ,"y":1.1},{"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]' * 2,
        schema={
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
        line_delimited=True,
    )
    assert (
        result.to_list()
        == [
            {"x": 1, "y": 1.1},
            {"x": 2, "y": 2.2},
            {"x": 3, "y": 3.3},
        ]
        * 2
    )
    assert str(result.type) == "6 * {x: int64, y: float64}"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * {x: int64, y: float64}"


def test_option_record():
    result = ak.operations.from_json(
        ' [ { "x" : 1 ,"y":1.1},null ,{"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]',
        schema={
            "type": "array",
            "items": {
                "type": ["object", "null"],
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.to_list() == [
        {"x": 1, "y": 1.1},
        None,
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]
    assert str(result.type) == "4 * ?{x: int64, y: float64}"

    result = ak.operations.from_json(
        ' [ { "x" : 1 ,"y":1.1},null ,{"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]' * 2,
        schema={
            "type": "array",
            "items": {
                "type": ["object", "null"],
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
        line_delimited=True,
    )
    assert (
        result.to_list()
        == [
            {"x": 1, "y": 1.1},
            None,
            {"x": 2, "y": 2.2},
            {"x": 3, "y": 3.3},
        ]
        * 2
    )
    assert str(result.type) == "8 * ?{x: int64, y: float64}"

    result = ak.operations.from_json(
        "[]",
        schema={
            "type": "array",
            "items": {
                "type": ["object", "null"],
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.to_list() == []
    assert str(result.type) == "0 * ?{x: int64, y: float64}"


def test_top_record():
    result = ak.operations.from_json(
        ' { "x" :1 ,"y":1.1}  ',
        schema={
            "type": "object",
            "properties": {"y": {"type": "number"}, "x": {"type": "integer"}},
            "required": ["x", "y"],
        },
    )
    assert result.to_list() == {"x": 1, "y": 1.1}

    result = ak.operations.from_json(
        ' { "x" :1 ,"y":1.1}  ' * 2,
        schema={
            "type": "object",
            "properties": {"y": {"type": "number"}, "x": {"type": "integer"}},
            "required": ["x", "y"],
        },
        line_delimited=True,
    )
    assert result.to_list() == [{"x": 1, "y": 1.1}] * 2

    result = ak.operations.from_json(
        "",
        schema={
            "type": "object",
            "properties": {"y": {"type": "number"}, "x": {"type": "integer"}},
            "required": ["x", "y"],
        },
        line_delimited=True,
    )
    assert result.to_list() == []


def test_number_substitutions():
    result = ak.operations.from_json(
        '[1, 2, 3.14, "nan", "-inf", "inf", 999]',
        schema={"type": "array", "items": {"type": "number"}},
        nan_string="nan",
        posinf_string="inf",
        neginf_string="-inf",
    )
    assert result.to_list()[:3] == [1, 2, 3.14]
    assert np.isnan(result[3])
    assert result.to_list()[4:] == [-np.inf, np.inf, 999]
    assert str(result.type) == "7 * float64"

    result = ak.operations.from_json(
        '["nan", "-inf", "inf"]',
        schema={"type": "array", "items": {"type": "number"}},
        nan_string="nan",
        posinf_string="inf",
        neginf_string="-inf",
    )
    assert np.isnan(result[0])
    assert result.to_list()[1:] == [-np.inf, np.inf]
    assert str(result.type) == "3 * float64"

    result = ak.operations.from_json(
        '["nan", "-inf", "inf"]',
        schema={"type": "array", "items": {"type": "string"}},
        nan_string="nan",
        posinf_string="inf",
        neginf_string="-inf",
    )
    assert result.to_list() == ["nan", "-inf", "inf"]
    assert str(result.type) == "3 * string"


def test_complex_substitutions():
    result = ak.operations.from_json(
        '[{"r": 1, "i": 1.1}, {"r": 2, "i": 2.2}, {"r": "inf", "i": 3.3}]',
        schema={
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"r": {"type": "number"}, "i": {"type": "number"}},
                "required": ["r", "i"],
            },
        },
        posinf_string="inf",
        complex_record_fields=("r", "i"),
    )
    assert result.to_list() == [1 + 1.1j, 2 + 2.2j, np.inf + 3.3j]
    assert str(result.type) == "3 * complex128"

    result = ak.operations.from_json(
        '[{"r": 1, "i": 1.1, "other": 1}, {"r": 2, "i": 2.2, "other": 1}, {"r": "inf", "i": 3.3, "other": 1}]',
        schema={
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "r": {"type": "number"},
                    "i": {"type": "number"},
                    "other": {"type": "integer"},
                },
                "required": ["r", "i", "other"],
            },
        },
        posinf_string="inf",
        complex_record_fields=("r", "i"),
    )
    assert result.to_list() == [1 + 1.1j, 2 + 2.2j, np.inf + 3.3j]
    assert str(result.type) == "3 * complex128"


def test_ignore_before():
    for what in [
        "null",
        "true",
        "2",
        "2.2",
        "[]",
        "[2]",
        "[2, 2.2]",
        "{}",
        '{"z": 2.2}',
        '{"z": []}',
        '{"z": [2]}',
        '{"z": [2, 2.2]}',
    ]:
        array = ak.from_json(
            '[{"y": ' + what + ', "x": 1}, {"x": 3}]',
            schema={
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {"x": {"type": "integer"}},
                    "required": ["x"],
                },
            },
        )
        assert array.to_list() == [{"x": 1}, {"x": 3}]
        assert str(array.type) == "2 * {x: int64}"


def test_ignore_after():
    for what in [
        "null",
        "true",
        "2",
        "2.2",
        "[]",
        "[2]",
        "[2, 2.2]",
        "{}",
        '{"z": 2.2}',
        '{"z": []}',
        '{"z": [2]}',
        '{"z": [2, 2.2]}',
    ]:
        array = ak.from_json(
            '[{"x": 1, "y": ' + what + '}, {"x": 3}]',
            schema={
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {"x": {"type": "integer"}},
                    "required": ["x"],
                },
            },
        )
        assert array.to_list() == [{"x": 1}, {"x": 3}]
        assert str(array.type) == "2 * {x: int64}"


def test_ignore_between():
    for what in [
        "null",
        "true",
        "2",
        "2.2",
        "[]",
        "[2]",
        "[2, 2.2]",
        "{}",
        '{"z": 2.2}',
        '{"z": []}',
        '{"z": [2]}',
        '{"z": [2, 2.2]}',
    ]:
        array = ak.from_json(
            '[{"x": 1, "y": ' + what + ', "z": true}, {"x": 3, "z": false}]',
            schema={
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {"z": {"type": "boolean"}, "x": {"type": "integer"}},
                    "required": ["z", "x"],
                },
            },
        )
        assert array.to_list() == [{"x": 1, "z": True}, {"x": 3, "z": False}]
        assert str(array.type) == "2 * {z: bool, x: int64}"


def test_option_ignore_before():
    for what in [
        "null",
        "true",
        "2",
        "2.2",
        "[]",
        "[2]",
        "[2, 2.2]",
        "{}",
        '{"z": 2.2}',
        '{"z": []}',
        '{"z": [2]}',
        '{"z": [2, 2.2]}',
    ]:
        array = ak.from_json(
            '[{"y": ' + what + ', "x": 1}, {"x": 3}]',
            schema={
                "type": "array",
                "items": {
                    "type": ["object", "null"],
                    "properties": {"x": {"type": "integer"}},
                    "required": ["x"],
                },
            },
        )
        assert array.to_list() == [{"x": 1}, {"x": 3}]
        assert str(array.type) == "2 * ?{x: int64}"


def test_option_ignore_after():
    for what in [
        "null",
        "true",
        "2",
        "2.2",
        "[]",
        "[2]",
        "[2, 2.2]",
        "{}",
        '{"z": 2.2}',
        '{"z": []}',
        '{"z": [2]}',
        '{"z": [2, 2.2]}',
    ]:
        array = ak.from_json(
            '[{"x": 1, "y": ' + what + '}, {"x": 3}]',
            schema={
                "type": "array",
                "items": {
                    "type": ["object", "null"],
                    "properties": {"x": {"type": "integer"}},
                    "required": ["x"],
                },
            },
        )
        assert array.to_list() == [{"x": 1}, {"x": 3}]
        assert str(array.type) == "2 * ?{x: int64}"


def test_option_ignore_between():
    for what in [
        "null",
        "true",
        "2",
        "2.2",
        "[]",
        "[2]",
        "[2, 2.2]",
        "{}",
        '{"z": 2.2}',
        '{"z": []}',
        '{"z": [2]}',
        '{"z": [2, 2.2]}',
    ]:
        array = ak.from_json(
            '[{"x": 1, "y": ' + what + ', "z": true}, {"x": 3, "z": false}]',
            schema={
                "type": "array",
                "items": {
                    "type": ["object", "null"],
                    "properties": {"z": {"type": "boolean"}, "x": {"type": "integer"}},
                    "required": ["z", "x"],
                },
            },
        )
        assert array.to_list() == [{"x": 1, "z": True}, {"x": 3, "z": False}]
        assert str(array.type) == "2 * ?{z: bool, x: int64}"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

numba = pytest.importorskip("numba")

import awkward.numba.layoutbuilder as lb  # noqa: E402

ak.numba.register_and_check()


def test_Numpy():
    builder = lb.Numpy(np.float64)

    builder.append(1.1)
    builder.append(2.2)
    builder.extend([3.3, 4.4, 5.5])

    error = ""
    assert builder.is_valid(error), error

    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.NumpyArray)
    assert str(ak.type(layout)) == "5 * float64"
    assert ak.to_list(layout) == [1.1, 2.2, 3.3, 4.4, 5.5]

    assert (
        str(builder.numbatype())
        == "ak.lb.Numpy(float64, parameters=Literal[NoneType](None))"
    )

    assert builder.form == layout.form

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 5
    builder.clear()
    assert len(builder) == 0

    assert builder.is_valid(error), error


def test_Numpy_char():
    builder = lb.Numpy(np.uint8, parameters={"__array__": "char"})
    builder.append(97)
    builder.append(98)
    builder.append(99)

    layout = builder.snapshot()
    assert str(ak.type(layout)) == "3 * char"
    assert ak.to_list(layout) == "abc"  # FIXME: ['a', 'b', 'c']????

    assert (
        str(builder.numbatype())
        == "ak.lb.Numpy(uint8, parameters=Literal[dict]({'__array__': 'char'}))"
    )

    error = ""
    assert builder.is_valid(error), error


def test_Empty():
    builder = lb.Empty()
    assert len(builder) == 0
    assert ak.to_list(builder.snapshot()) == []

    error = ""
    assert builder.is_valid(error), error

    with pytest.raises(AttributeError):
        builder.content.append(1.1)

    with pytest.raises(AttributeError):
        builder.content.extend([3.3, 4.4, 5.5])

    error = ""
    assert builder.is_valid(error), error

    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.EmptyArray)
    assert str(ak.type(layout)) == "0 * unknown"
    assert ak.to_list(layout) == []

    assert str(builder.numbatype()) == "ak.lb.Empty(parameters=Literal[NoneType](None))"

    builder = lb.Empty()
    assert str(builder.numbatype()) == "ak.lb.Empty(parameters=Literal[NoneType](None))"
    assert builder.form == layout.form


def test_ListOffset():
    builder = lb.ListOffset(np.int32, lb.Numpy(np.float64))
    assert len(builder) == 0
    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.ListOffsetArray)
    assert ak.to_list(layout) == []

    content = builder.begin_list()
    content.append(1.1)
    content.append(2.2)
    content.append(3.3)
    builder.end_list()

    builder.begin_list()
    builder.end_list()

    builder.begin_list()
    content.append(4.4)
    content.append(5.5)
    builder.end_list()

    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.ListOffsetArray)
    assert ak.to_list(layout) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert builder.form == layout.form

    error = ""
    assert builder.is_valid(error), error

    assert (
        str(builder.numbatype())
        == "ak.lb.ListOffset(int32, ak.lb.Numpy(float64, parameters=Literal[NoneType](None)), parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_Regular():
    builder = lb.Regular(lb.Numpy(np.float64), 3)
    assert len(builder) == 0
    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.RegularArray)
    assert ak.to_list(layout) == []

    content = builder.begin_list()
    content.append(1.1)
    content.append(2.2)
    content.append(3.3)
    builder.end_list()

    builder.begin_list()
    content.append(4.4)
    content.append(5.5)
    content.append(6.6)
    builder.end_list()

    layout = builder.snapshot()
    assert ak.to_list(layout) == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]
    assert len(builder) == 2
    assert builder.form == layout.form

    error = ""
    assert builder.is_valid(error), error

    assert (
        str(builder.numbatype())
        == "ak.lb.Regular(ak.lb.Numpy(float64, parameters=Literal[NoneType](None)), 3, parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_Regular_size0():
    with pytest.raises(ValueError):
        lb.Regular(lb.Numpy(np.float64), 0)


def test_IndexedOption():
    builder = lb.IndexedOption(np.int64, lb.Numpy(np.float64))
    assert len(builder) == 0

    content = builder.append_valid()
    content.append(1.1)
    builder.append_invalid()

    data = np.array([3.3, 4.4, 5.5], dtype=np.float64)
    builder.extend_valid(len(data))
    content.extend(data)

    builder.extend_invalid(2)
    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1, None, 3.3, 4.4, 5.5, None, None]
    assert builder.form == layout.form

    assert len(builder) == 7

    error = ""
    assert builder.is_valid(error), error

    assert (
        str(builder.numbatype())
        == "ak.lb.IndexedOption(int64, ak.lb.Numpy(float64, parameters=Literal[NoneType](None)), parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_Record():
    builder = lb.Record(
        [
            lb.Numpy(np.float64),
            lb.Numpy(np.int64),
            lb.Numpy(np.uint8, parameters={"__array__": "char"}),
        ],
        ["one", "two", "three"],
    )
    assert len(builder) == 0

    one = builder.content("one")
    two = builder.content("two")
    three = builder.content("three")

    three.append(0x61)  #'a')

    one.append(1.1)
    one.append(3.3)

    two.append(2)
    two.append(4)

    three.append(0x62)  #'b')

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        {"one": 1.1, "two": 2, "three": "a"},
        {"one": 3.3, "two": 4, "three": "b"},
    ]
    assert builder.form == layout.form

    assert len(builder) == 2

    error = ""
    assert builder.is_valid(error), error

    assert (
        str(builder.numbatype())
        == "ak.lb.Record((ak.lb.Numpy(float64, parameters=Literal[NoneType](None)), ak.lb.Numpy(int64, parameters=Literal[NoneType](None)), ak.lb.Numpy(uint8, parameters=Literal[dict]({'__array__': 'char'}))), ('one', 'two', 'three'), parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_IndexedOption_Record():
    builder = lb.IndexedOption(
        np.int64, lb.Record([lb.Numpy(np.float64), lb.Numpy(np.int64)], ["x", "y"])
    )
    assert len(builder) == 0
    record = builder.append_valid()
    x = record.content("x")
    y = record.content("y")

    x.append(1.1)
    y.append(2)

    builder.append_invalid()

    builder.append_valid()
    x.append(3.3)
    y.append(4)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        {"x": 1.1, "y": 2},
        None,
        {"x": 3.3, "y": 4},
    ]
    assert builder.form == layout.form


def test_Tuple_Numpy_ListOffset():
    builder = lb.Tuple(
        [lb.Numpy(np.float64), lb.ListOffset(np.int64, lb.Numpy(np.int32))]
    )
    assert len(builder) == 0

    error = ""
    assert builder.is_valid(error) is True

    one = builder.index(0)
    one.append(1.1)
    two = builder.index(1)
    two_list = two.begin_list()
    two_list.append(1)
    two.end_list()

    assert builder.is_valid(error) is True

    one.append(2.2)
    two.begin_list()
    two_list.append(1)
    two_list.append(2)
    two.end_list()

    assert builder.is_valid(error) is True

    one.append(3.3)
    two.begin_list()
    two_list.append(1)
    two_list.append(2)
    two_list.append(3)
    two.end_list()

    layout = builder.snapshot()
    assert ak.to_list(layout) == [(1.1, [1]), (2.2, [1, 2]), (3.3, [1, 2, 3])]

    assert builder.is_valid(error) is True
    assert builder.form == layout.form


def test_Unmasked():
    builder = lb.Unmasked(lb.Numpy(np.int64))
    assert len(builder) == 0

    content = builder._content
    content.append(11)
    content.append(22)
    content.append(33)
    content.append(44)
    content.append(55)

    err = ""
    assert builder.is_valid(err) is True

    layout = builder.snapshot()
    assert ak.to_list(layout) == [11, 22, 33, 44, 55]
    assert builder.form == layout.form

    assert (
        str(builder.numbatype())
        == "ak.lb.Unmasked(ak.lb.Numpy(int64, parameters=Literal[NoneType](None)), parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_ByteMasked():
    builder = lb.ByteMasked(lb.Numpy(np.float64), valid_when=True)
    assert len(builder) == 0

    content = builder.append_valid()
    content.append(1.1)

    builder.append_invalid()
    content.append(np.nan)

    data = np.array([3.3, 4.4, 5.5], dtype=np.float64)
    builder.extend_valid(3)
    content.extend(data)

    builder.extend_invalid(2)
    content.append(np.nan)
    content.append(np.nan)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1, None, 3.3, 4.4, 5.5, None, None]
    assert builder.form == layout.form

    error = ""
    assert builder.is_valid(error), error

    assert (
        str(builder.numbatype())
        == "ak.lb.ByteMasked(ak.lb.Numpy(float64, parameters=Literal[NoneType](None)), valid_when=True, parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_BitMasked():
    builder = lb.BitMasked(np.uint8, lb.Numpy(np.float64), True, True)
    assert len(builder) == 0

    subbuilder = builder.append_valid()
    subbuilder.append(1.1)
    assert len(builder) == 1
    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1]
    assert builder.form == layout.form

    builder.append_invalid()
    subbuilder.append(np.nan)
    assert len(builder) == 2

    data = np.array([3.3, 4.4, 5.5], dtype=np.float64)

    builder.extend_valid(3)
    subbuilder.extend(data)
    assert len(builder) == 5

    builder.extend_invalid(2)
    subbuilder.append(np.nan)
    subbuilder.append(np.nan)

    assert len(builder) == 7

    builder.append_valid()
    subbuilder.append(8)
    assert len(builder) == 8

    builder.append_valid()
    subbuilder.append(9)
    assert len(builder) == 9

    builder.append_valid()
    subbuilder.append(10)
    assert len(builder) == 10

    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1, None, 3.3, 4.4, 5.5, None, None, 8, 9, 10]

    error = ""
    assert builder.is_valid(error), error

    assert (
        str(builder.numbatype())
        == "ak.lb.BitMasked(uint8, ak.lb.Numpy(float64, parameters=Literal[NoneType](None)), True, True, parameters=Literal[NoneType](None))"
    )
    builder.clear()
    assert len(builder) == 0


def test_Union_Numpy_ListOffset():
    builder = lb.Union(
        np.int8,
        np.int64,
        [
            lb.Numpy(np.float64),
            lb.ListOffset(np.int64, lb.Numpy(np.int32)),
        ],
    )
    assert len(builder) == 0

    # error = ""
    # assert builder.is_valid(error) == True

    one = builder.append_content(0)
    one.append(1.1)

    # assert builder.is_valid(error) == True

    two = builder.append_content(1)
    list = two.begin_list()
    list.append(1)
    list.append(2)
    two.end_list()

    # assert builder.is_valid(error) == True

    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1, [1, 2]]
    assert builder.form == layout.form


def test_Union_ListOffset_Record():
    builder = lb.Union(
        np.int8,
        np.int64,
        [
            lb.ListOffset(np.int64, lb.Numpy(np.int32)),
            lb.Record([lb.Numpy(np.float64), lb.Numpy(np.int64)], ["x", "y"]),
        ],
    )
    assert len(builder) == 0

    one = builder.append_content(0)
    list = one.begin_list()
    list.append(1)
    list.append(3)
    one.end_list()

    two = builder.append_content(1)
    x = two.content("x")
    y = two.content("y")

    x.append(1.1)
    y.append(11)

    builder.append_content(0)
    list = one.begin_list()
    list.append(5.5)
    one.end_list()

    builder.append_content(1)
    x.append(2.2)
    y.append(22)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [[1, 3], {"x": 1.1, "y": 11}, [5], {"x": 2.2, "y": 22}]
    assert builder.form == layout.form


def test_unbox():
    @numba.njit(debug=True)
    def f1(x):
        x  # noqa: B018 (we want to test the unboxing)
        return 3.14

    builder = lb.BitMasked(np.uint8, lb.Numpy(np.float64), True, True)
    f1(builder)

    builder = lb.ByteMasked(lb.Numpy(np.float64), valid_when=True)
    f1(builder)

    builder = lb.Empty()
    f1(builder)

    builder = lb.IndexedOption(np.int64, lb.Numpy(np.float64))
    f1(builder)

    builder = lb.IndexedOption(np.int32, lb.Numpy(np.float64))
    f1(builder)

    builder = lb.ListOffset(np.int32, lb.ListOffset(np.int64, lb.Numpy(np.int64)))
    f1(builder)

    builder = lb.ListOffset(np.int32, lb.Numpy(np.float64))
    f1(builder)

    builder = lb.ListOffset(np.int32, lb.Empty())
    f1(builder)

    builder = lb.ListOffset(np.int8, lb.Numpy(np.int64))
    f1(builder)

    builder = lb.ListOffset(np.int32, lb.Empty())
    f1(builder)

    builder = lb.Numpy(np.int32)
    f1(builder)

    builder = lb.Record(
        [
            lb.Numpy(np.float64),
            lb.Numpy(np.int64),
            lb.Numpy(np.uint8),
        ],
        ["one", "two", "three"],
    )
    f1(builder)

    @numba.njit(debug=True)
    def f_compiled_index(x):
        i = x._contents[0]
        i = x._contents[1]
        i = x._contents[2]
        return i

    f_compiled_index(builder)

    @numba.njit(debug=True)
    def f_runtime_index(x, name):
        return x.content("one")  # name) # FIXME: numba.types.UnicodeType

    content = f_runtime_index(builder, "one")
    assert content.numbatype() == builder._contents[0].numbatype()

    builder = lb.Regular(lb.Numpy(np.float64), size=3)
    f1(builder)

    builder = lb.Tuple(
        [lb.Numpy(np.float64), lb.ListOffset(np.int64, lb.Numpy(np.int32))]
    )
    f1(builder)

    builder = lb.Union(
        np.int8,
        np.int64,
        [
            lb.Numpy(np.float64),
            lb.ListOffset(np.int64, lb.Numpy(np.int32)),
        ],
    )
    f1(builder)

    builder = lb.Unmasked(lb.Numpy(np.int64))
    f1(builder)


def test_unbox_for_loop():
    @numba.njit
    def f2(x):
        for i in range(0, 10):
            x.append(i)
        return

    builder = lb.Numpy(np.int64, parameters=None, initial=10, resize=2.0)
    f2(builder)
    assert ak.to_list(builder.snapshot()) == list(range(10))

    builder = lb.Empty()
    # Unknown attribute 'append' of type lb.Empty
    with pytest.raises(numba.core.errors.TypingError):
        f2(builder)


def test_box():
    @numba.njit
    def f3(x):
        return x

    builder = lb.Numpy(np.int32)
    out1 = f3(builder)
    assert ak.to_list(out1.snapshot()) == []

    for x in range(15):
        out1.append(x)

    out2 = f3(out1)
    assert ak.to_list(out2.snapshot()) == list(range(15))

    builder = lb.Empty()
    out3 = f3(builder)
    assert ak.to_list(out3.snapshot()) == []

    builder = lb.ListOffset(np.int64, lb.Numpy(np.int32))
    out4 = f3(builder)
    assert ak.to_list(out4.snapshot()) == []

    builder = lb.ListOffset(np.int64, lb.Numpy(np.int64))
    out5 = f3(builder)
    assert ak.to_list(out5.snapshot()) == []

    builder = lb.ListOffset(np.int32, lb.Empty())
    out6 = f3(builder)
    assert ak.to_list(out6.snapshot()) == []

    builder = lb.ListOffset(np.int32, lb.ListOffset(np.int64, lb.Numpy(np.int64)))
    out5 = f3(builder)
    assert ak.to_list(out5.snapshot()) == []

    builder = lb.Regular(lb.Numpy(np.float64), 3)
    out7 = f3(builder)
    assert ak.to_list(out7.snapshot()) == []

    builder = lb.ListOffset(np.int32, lb.Regular(lb.Numpy(np.float64), 3))
    out8 = f3(builder)
    assert ak.to_list(out8.snapshot()) == []

    builder = lb.IndexedOption(np.int32, lb.Numpy(np.float64))
    out9 = f3(builder)
    assert ak.to_list(out9.snapshot()) == []

    builder = lb.IndexedOption(np.int64, lb.Numpy(np.float64))
    out10 = f3(builder)
    assert ak.to_list(out10.snapshot()) == []

    builder = lb.ByteMasked(lb.Numpy(np.float64), valid_when=True)
    out11 = f3(builder)
    assert ak.to_list(out11.snapshot()) == []

    builder = lb.BitMasked(np.uint8, lb.Numpy(np.float64), True, True)
    out12 = f3(builder)
    assert ak.to_list(out12.snapshot()) == []

    builder = lb.Unmasked(lb.Numpy(np.int64))
    out13 = f3(builder)
    assert ak.to_list(out13.snapshot()) == []

    builder = lb.Record(
        [
            lb.Numpy(np.float64),
            lb.Numpy(np.int64),
            lb.Numpy(np.uint8),
        ],
        ["one", "two", "three"],
    )
    out14 = f3(builder)
    assert ak.to_list(out14.snapshot()) == []

    builder = lb.Tuple(
        [lb.Numpy(np.float64), lb.ListOffset(np.int64, lb.Numpy(np.int32))]
    )
    out15 = f3(builder)
    assert ak.to_list(out15.snapshot()) == []

    builder = lb.Union(
        np.int8,
        np.int64,
        [
            lb.Numpy(np.float64),
            lb.ListOffset(np.int64, lb.Numpy(np.int32)),
        ],
    )
    out16 = f3(builder)
    assert ak.to_list(out16.snapshot()) == []


def test_len():
    @numba.njit
    def f4(x):
        return len(x)

    builder = lb.Numpy(np.int32, parameters=None, initial=10, resize=2.0)
    assert f4(builder) == 0

    builder.append(123)
    assert f4(builder) == 1

    builder = lb.Empty()
    assert f4(builder) == 0

    builder = lb.ListOffset(np.int64, lb.Numpy(np.int32))
    assert f4(builder) == 0

    builder = lb.ListOffset(np.int8, lb.Empty())
    assert f4(builder) == 0

    builder = lb.ListOffset(np.int32, lb.ListOffset(np.int32, lb.Numpy(np.int64)))
    assert f4(builder) == 0

    builder = lb.ListOffset(np.int64, lb.Numpy(np.int32))
    assert f4(builder) == 0

    builder = lb.Regular(lb.Numpy(np.float64), 3)
    assert f4(builder) == 0

    builder = lb.IndexedOption(np.int32, lb.Numpy(np.float64))
    assert f4(builder) == 0

    builder = lb.IndexedOption(np.int64, lb.Numpy(np.float64))
    assert f4(builder) == 0

    builder = lb.ByteMasked(lb.Numpy(np.float64), valid_when=True)
    assert f4(builder) == 0

    builder = lb.BitMasked(np.uint8, lb.Numpy(np.float64), True, True)
    assert f4(builder) == 0

    builder = lb.Unmasked(lb.Numpy(np.int64))
    assert f4(builder) == 0

    builder = lb.Record(
        [
            lb.Numpy(np.float64),
            lb.Numpy(np.int64),
            lb.Numpy(np.uint8),
        ],
        ["one", "two", "three"],
    )
    assert f4(builder) == 0

    builder = lb.Tuple(
        [lb.Numpy(np.float64), lb.ListOffset(np.int64, lb.Numpy(np.int32))]
    )
    assert f4(builder) == 0

    builder = lb.Union(
        np.int8,
        np.int64,
        [
            lb.Numpy(np.float64),
            lb.ListOffset(np.int64, lb.Numpy(np.int32)),
        ],
    )
    assert f4(builder) == 0


def test_Numpy_from_buffer():
    @numba.njit
    def f5(debug=True):
        growablebuffer = ak.numba.GrowableBuffer(np.float64)
        growablebuffer.append(66.6)
        growablebuffer.append(77.7)
        return growablebuffer

    out = f5()
    assert out.snapshot().tolist() == [66.6, 77.7]

    @numba.njit
    def f6():
        growablebuffer = ak.numba.GrowableBuffer(np.float64)
        growablebuffer.append(66.6)
        growablebuffer.append(77.7)

        return ak._connect.numba.layoutbuilder._from_buffer(growablebuffer)

    out = f6()
    assert isinstance(out, lb.Numpy)
    assert out.dtype == np.dtype(np.float64)
    assert len(out) == 2

    assert ak.to_list(out.snapshot()) == [66.6, 77.7]


def test_Numpy_ctor():
    @numba.njit
    def f7():
        return lb.Numpy("f4")

    out = f7()
    assert isinstance(out, lb.Numpy)
    assert out.dtype == np.dtype("f4")
    assert len(out) == 0

    @numba.njit
    def f8():
        return lb.Numpy(np.float32)

    out = f8()
    assert isinstance(out, lb.Numpy)
    assert out.dtype == np.dtype(np.float32)
    assert len(out) == 0

    @numba.njit
    def f9():
        return lb.Numpy(np.dtype(np.float32))

    out = f9()
    assert isinstance(out, lb.Numpy)
    assert out.dtype == np.dtype(np.float32)
    assert len(out) == 0


def test_Numpy_append():
    @numba.njit
    def f10(builder):
        for i in range(8):
            builder.append(i)

    builder = lb.Numpy(np.float32)

    f10(builder)

    assert ak.to_list(builder.snapshot()) == list(range(8))

    f10(builder)

    assert ak.to_list(builder.snapshot()) == list(range(8)) + list(range(8))


def test_Numpy_extend():
    @numba.njit
    def f11(builder):
        builder.extend(np.arange(8))

    builder = lb.Numpy(np.float32)

    f11(builder)

    assert ak.to_list(builder.snapshot()) == list(range(8))

    f11(builder)

    assert ak.to_list(builder.snapshot()) == list(range(8)) + list(range(8))


def test_ListOffset_begin_list():
    @numba.njit
    def f13(builder):
        return builder.begin_list()

    builder = lb.ListOffset(np.int64, lb.Numpy(np.int32))

    out = f13(builder)
    assert isinstance(out, lb.Numpy)
    assert out.dtype == np.dtype(np.int32)


def test_ListOffset_end_list():
    @numba.njit
    def f14(builder):
        builder.begin_list()
        builder.end_list()

        builder.begin_list()
        builder.end_list()

        builder.begin_list()
        builder.end_list()

    builder = lb.ListOffset(np.int64, lb.Numpy(np.float64))
    assert len(builder) == 0

    f14(builder)
    assert len(builder) == 3

    assert ak.to_list(builder.snapshot()) == [[], [], []]


def test_ListOffset_append():
    @numba.njit
    def f15(builder):
        content = builder.begin_list()
        content.append(1.1)
        content.append(2.2)
        content.append(3.3)
        builder.end_list()

        builder.begin_list()
        builder.end_list()

        builder.begin_list()
        content.append(4.4)
        content.append(5.5)
        builder.end_list()

    builder = lb.ListOffset(np.int64, lb.Numpy(np.float64))
    assert len(builder) == 0

    f15(builder)
    assert len(builder) == 3

    assert ak.to_list(builder.snapshot()) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_ListOffset_as_string():
    @numba.njit
    def f16(builder):
        content = builder.begin_list()
        content.append(104)  # 'h'
        content.append(101)  # 'e'
        content.append(108)  # 'l'
        content.append(108)  # 'l'
        content.append(111)  # 'o'
        builder.end_list()

        builder.begin_list()
        content.append(119)  # 'w'
        content.append(111)  # 'o'
        content.append(114)  # 'r'
        content.append(108)  # 'l'
        content.append(100)  # 'd'
        builder.end_list()

    builder = lb.ListOffset(
        np.int64,
        lb.Numpy(np.uint8, parameters={"__array__": "char"}),
        parameters={"__array__": "string"},
    )
    f16(builder)

    assert ak.to_list(builder.snapshot()) == ["hello", "world"]


def test_ListOffset_append2():
    @numba.njit
    def f17(builder):
        content = builder.begin_list()
        content.append(1.1)
        content.append(2.2)
        content.append(3.3)
        builder.end_list()

        builder.begin_list()
        builder.end_list()

        builder.begin_list()
        content.append(4.4)
        content.append(5.5)
        builder.end_list()

    builder = lb.ListOffset(np.int32, lb.Numpy(np.float64))
    assert len(builder) == 0

    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.ListOffsetArray)
    assert ak.to_list(layout) == []

    f17(builder)

    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.ListOffsetArray)
    assert ak.to_list(layout) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 3
    builder.clear()
    assert len(builder) == 0


def test_Regular_append():
    @numba.njit
    def f18(builder):
        content = builder.begin_list()
        content.append(1.1)
        content.append(2.2)
        content.append(3.3)
        builder.end_list()

        builder.begin_list()
        content.append(4.4)
        content.append(5.5)
        content.append(6.6)
        builder.end_list()

    builder = lb.Regular(lb.Numpy(np.float64), 3)
    assert len(builder) == 0
    layout = builder.snapshot()
    assert isinstance(layout, ak.contents.RegularArray)
    assert ak.to_list(layout) == []

    f18(builder)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]

    error = ""
    assert builder.is_valid(error), error


def test_IndexedOption_Record_append():
    @numba.njit
    def f19(builder):
        record = builder.append_valid()
        x = record.content("x")
        y = record.content("y")
        x.append(1.1)
        y.append(2)
        builder.append_invalid()

        builder.append_valid()
        x.append(3.3)
        y.append(4)

    builder = lb.IndexedOption(
        np.int64, lb.Record([lb.Numpy(np.float64), lb.Numpy(np.int64)], ["x", "y"])
    )
    assert len(builder) == 0

    f19(builder)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        {"x": 1.1, "y": 2},
        None,
        {"x": 3.3, "y": 4},
    ]
    assert builder.form == layout.form


def test_IndexedOption_append_extend():
    @numba.njit
    def f21(builder):
        content = builder.append_valid()
        content.append(1.1)
        builder.append_valid()
        content.append(2.2)
        builder.append_valid()
        content.append(3.3)
        builder.append_valid()
        content.append(4.4)
        builder.append_valid()
        content.append(5.5)
        builder.append_valid()
        content.append(6.6)

    @numba.njit
    def f22(builder):
        builder.append_invalid()

    @numba.njit
    def f23(builder, data):
        content = builder.extend_valid(len(data))
        content.extend(data)

    @numba.njit
    def f24(builder, size):
        builder.extend_invalid(size)

    builder = lb.IndexedOption(np.int64, lb.Numpy(np.float64))
    assert len(builder) == 0

    f21(builder)
    assert len(builder) == 6

    f22(builder)
    assert len(builder) == 7

    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, None]

    # FIXME: keep track of self._last_valid
    # error = ""
    # assert builder.is_valid(error), error

    data = np.array([33.33, 44.44, 55.55], dtype=np.float64)
    f23(builder, data)
    assert len(builder) == 10

    f24(builder, 3)
    assert len(builder) == 13

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        None,
        33.33,
        44.44,
        55.55,
        None,
        None,
        None,
    ]

    # FIXME: assert builder.is_valid(error), error


def test_ByteMasked_append_extend():
    @numba.njit
    def f25(builder):
        content = builder.append_valid()
        content.append(1.1)
        builder.append_valid()
        content.append(2.2)
        builder.append_valid()
        content.append(3.3)
        builder.append_valid()
        content.append(4.4)
        builder.append_valid()
        content.append(5.5)
        builder.append_valid()
        content.append(6.6)

    @numba.njit
    def f26(builder):
        content = builder.append_invalid()
        content.append(np.nan)

    @numba.njit
    def f27(builder, data):
        content = builder.extend_valid(len(data))
        content.extend(data)

    @numba.njit
    def f28(builder, size):
        content = builder.extend_invalid(size)
        content.extend([np.nan] * size)

    builder = lb.ByteMasked(lb.Numpy(np.float64), valid_when=True)
    assert len(builder) == 0

    f25(builder)
    f26(builder)
    data = np.array([3.3, 4.4, 5.5], dtype=np.float64)
    f27(builder, data)

    f28(builder, 2)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        None,
        3.3,
        4.4,
        5.5,
        None,
        None,
    ]

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 12
    builder.clear()
    assert len(builder) == 0


def test_BitMasked_append_extend():
    @numba.njit
    def f29(builder):
        content = builder.append_valid()
        content.append(1.1)
        builder.append_valid()
        content.append(2.2)
        builder.append_valid()
        content.append(3.3)
        builder.append_valid()
        content.append(4.4)
        builder.append_valid()
        content.append(5.5)
        builder.append_valid()
        content.append(6.6)

    @numba.njit
    def f30(builder):
        content = builder.append_invalid()
        content.append(np.nan)

    @numba.njit
    def f31(builder, data):
        content = builder.extend_valid(len(data))
        content.extend(data)

    @numba.njit
    def f32(builder, size):
        content = builder.extend_invalid(size)
        content.append(np.nan)
        content.append(np.nan)

    builder = lb.BitMasked(np.uint8, lb.Numpy(np.float64), True, True)
    assert len(builder) == 0

    f29(builder)
    f30(builder)
    data = np.array([3.3, 4.4, 5.5], dtype=np.float64)
    f31(builder, data)

    f32(builder, 2)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        None,
        3.3,
        4.4,
        5.5,
        None,
        None,
    ]

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 12
    builder.clear()
    assert len(builder) == 0


def test_Unmasked_append_extend():
    @numba.njit
    def f33(builder):
        content = builder._content
        content.append(1.1)
        content.append(2.2)
        content.append(3.3)
        content.append(4.4)
        content.append(5.5)
        content.append(6.6)

    @numba.njit
    def f34(builder, data):
        content = builder._content
        content.extend(data)

    builder = lb.Unmasked(lb.Numpy(np.float64))

    f33(builder)
    data = np.array([3.3, 4.4, 5.5], dtype=np.float64)
    f34(builder, data)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        3.3,
        4.4,
        5.5,
    ]

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 9
    builder.clear()
    assert len(builder) == 0


def test_Record_content():
    @numba.njit
    def field_index(builder, name):
        return builder._field_index(name)

    @numba.njit
    def fill(builder):
        content_one = builder.content("one")
        content_one.append(1.1)
        content_two = builder.content("two")
        content_two.append(1)
        content_three = builder.content("three")
        content_three.append(111)

    builder = lb.Record(
        [
            lb.Numpy(np.float64),
            lb.Numpy(np.int64),
            lb.Numpy(np.uint8, parameters={"__array__": "char"}),
        ],
        ["one", "two", "three"],
    )
    assert field_index(builder, "one") == 0
    assert field_index(builder, "two") == 1
    assert field_index(builder, "three") == 2

    with pytest.raises(ValueError):
        field_index(builder, "four")  # ValueError: tuple.index(x): x not in tuple

    fill(builder)
    layout = builder.snapshot()
    assert ak.to_list(layout) == [{"one": 1.1, "three": "o", "two": 1}]  # ???
    assert builder.form == layout.form

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 1
    builder.clear()
    assert len(builder) == 0


def test_Tuple_append():
    @numba.njit
    def f38(builder):
        content_one = builder.content(0)
        content_one.append(1.1)
        content_two = builder.content(1)
        content_list = content_two.begin_list()
        content_list.append(1)
        content_list.append(2)
        content_list.append(3)
        content_two.end_list()

    builder = lb.Tuple(
        [lb.Numpy(np.float64), lb.ListOffset(np.int64, lb.Numpy(np.int32))]
    )
    f38(builder)
    layout = builder.snapshot()
    assert ak.to_list(layout) == [(1.1, [1, 2, 3])]

    error = ""
    assert builder.is_valid(error), error

    assert len(builder) == 1
    builder.clear()
    assert len(builder) == 0


def test_Union_append():
    @numba.njit
    def f39(builder):
        one = builder.append_content(0)
        one.append(1.1)

        two = builder.append_content(1)
        list = two.begin_list()
        list.append(1)
        list.append(2)
        list.append(3)
        two.end_list()

    builder = lb.Union(
        np.int8,
        np.int64,
        [
            lb.Numpy(np.float64),
            lb.ListOffset(np.int64, lb.Numpy(np.int32)),
        ],
    )
    assert len(builder) == 0

    f39(builder)

    layout = builder.snapshot()
    assert ak.to_list(layout) == [1.1, [1, 2, 3]]


def test_numba_append():
    @numba.njit
    def create():
        return lb.Numpy(np.int32)

    @numba.njit
    def append_range(builder, start, stop):
        for x in range(start, stop):
            builder.append(x)

    @numba.njit
    def append_single(builder, x):
        builder.append(x)

    builder = create()
    assert ak.to_list(builder.snapshot()) == []
    assert len(builder) == 0

    append_range(builder, 0, 5)
    assert ak.to_list(builder.snapshot()) == list(range(5))
    assert len(builder) == 5

    append_range(builder, 5, 9)
    assert ak.to_list(builder.snapshot()) == list(range(9))
    assert len(builder) == 9

    append_single(builder, 9)
    assert ak.to_list(builder.snapshot()) == list(range(10))
    assert len(builder) == 10

    append_single(builder, 10)
    assert ak.to_list(builder.snapshot()) == list(range(11))
    assert len(builder) == 11

    append_single(builder, 11)
    assert ak.to_list(builder.snapshot()) == list(range(12))
    assert len(builder) == 12

    append_range(builder, 12, 30)
    assert ak.to_list(builder.snapshot()) == list(range(30))
    assert len(builder) == 30

    append_single(builder, 30)
    assert ak.to_list(builder.snapshot()) == list(range(31))
    assert len(builder) == 31



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list

"""includes test_0117, test_0110, test_0042, test_0127,
 test_0198, test_0446, test_0585, test_0590, test_0612,
 test_0724, test_0866, test_0973
"""


def test_flatten_ListOffsetArray():
    array = ak.highlevel.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout

    assert ak.operations.to_list(ak.operations.flatten(array)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:])) == [4.4, 5.5]

    array = ak.highlevel.Array(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[], [6.6, 7.7, 8.8, 9.9]]]
    ).layout

    assert ak.operations.to_list(ak.operations.flatten(array)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:])) == [
        [5.5],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:])) == [
        [],
        [3.3, 4.4],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:], axis=2)) == [
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        [],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]

    array = ak.highlevel.Array(
        np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7).tolist()
    ).layout

    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=1))
        == np.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=2))
        == np.arange(2 * 3 * 5 * 7).reshape(2, 3 * 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=3))
        == np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5 * 7).tolist()
    )

    array = ak.highlevel.Array(
        ak.operations.from_iter(
            np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7).tolist(), highlevel=False
        )
    ).layout

    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=1))
        == np.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=2))
        == np.arange(2 * 3 * 5 * 7).reshape(2, 3 * 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=3))
        == np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5 * 7).tolist()
    )

    array = ak.highlevel.Array(np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7)).layout

    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=1))
        == np.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=2))
        == np.arange(2 * 3 * 5 * 7).reshape(2, 3 * 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(array, axis=3))
        == np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5 * 7).tolist()
    )


def test_flatten_IndexedArray():
    array = ak.highlevel.Array(
        [[1.1, 2.2, None, 3.3], None, [], None, [4.4, 5.5], None]
    ).layout

    assert ak.operations.to_list(ak.operations.flatten(array)) == [
        1.1,
        2.2,
        None,
        3.3,
        4.4,
        5.5,
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:])) == [4.4, 5.5]

    array = ak.highlevel.Array(
        [
            [[0.0, 1.1, 2.2], None, None, [3.3, 4.4]],
            [],
            [[5.5]],
            [[], [6.6, 7.7, 8.8, 9.9]],
        ]
    ).layout

    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:], axis=2)) == [
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        [],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]

    array = ak.highlevel.Array(
        [
            [[0.0, 1.1, 2.2], [3.3, 4.4]],
            [],
            [[5.5]],
            None,
            None,
            [[], [6.6, 7.7, 8.8, 9.9]],
        ]
    ).layout

    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        [5.5],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:], axis=2)) == [
        [],
        [5.5],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        [],
        [],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]

    array = ak.highlevel.Array(
        [
            [[0.0, 1.1, None, 2.2], None, [], None, [3.3, 4.4]],
            None,
            [],
            [[5.5]],
            None,
            [[], [6.6, None, 7.7, 8.8, 9.9], None],
        ]
    ).layout

    assert ak.operations.to_list(ak.operations.flatten(array)) == [
        [0.0, 1.1, None, 2.2],
        None,
        [],
        None,
        [3.3, 4.4],
        [5.5],
        [],
        [6.6, None, 7.7, 8.8, 9.9],
        None,
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [0.0, 1.1, None, 2.2, 3.3, 4.4],
        None,
        [],
        [5.5],
        None,
        [6.6, None, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:], axis=2)) == [
        None,
        [],
        [5.5],
        None,
        [6.6, None, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        None,
        [],
        [],
        None,
        [6.6, None, 7.7, 8.8, 9.9],
    ]

    content = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.index.Index64(np.array([2, 1, 0, 3, 3, 4], dtype=np.int64))
    array = ak.contents.IndexedArray(index, content)

    assert to_list(array) == [
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        [5.5],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array)) == [
        3.3,
        4.4,
        0.0,
        1.1,
        2.2,
        5.5,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]

    content = ak.operations.from_iter(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[], [6.6, 7.7, 8.8, 9.9]]],
        highlevel=False,
    )
    index = ak.index.Index64(np.array([2, 2, 1, 0, 3], dtype=np.int64))
    array = ak.contents.IndexedArray(index, content)

    assert to_list(array) == [
        [[5.5]],
        [[5.5]],
        [],
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [[], [6.6, 7.7, 8.8, 9.9]],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [5.5],
        [5.5],
        [],
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [6.6, 7.7, 8.8, 9.9],
    ]


def test_flatten_RecordArray():
    array = ak.highlevel.Array(
        [
            {"x": [], "y": [[3, 3, 3]]},
            {"x": [[1]], "y": [[2, 2]]},
            {"x": [[2], [2]], "y": [[1]]},
            {"x": [[3], [3], [3]], "y": [[]]},
        ]
    ).layout

    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        {"x": [], "y": [3, 3, 3]},
        {"x": [1], "y": [2, 2]},
        {"x": [2, 2], "y": [1]},
        {"x": [3, 3, 3], "y": []},
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:], axis=2)) == [
        {"x": [1], "y": [2, 2]},
        {"x": [2, 2], "y": [1]},
        {"x": [3, 3, 3], "y": []},
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:], axis=2)) == [
        {"x": [], "y": []},
        {"x": [], "y": []},
        {"x": [2], "y": []},
        {"x": [3, 3], "y": []},
    ]


def test_flatten_UnionArray():
    content1 = ak.operations.from_iter(
        [[1.1], [2.2, 2.2], [3.3, 3.3, 3.3]], highlevel=False
    )
    content2 = ak.operations.from_iter(
        [[[3, 3, 3], [3, 3, 3], [3, 3, 3]], [[2, 2], [2, 2]], [[1]]], highlevel=False
    )
    content3 = ak.operations.from_iter(
        [
            [["3", "3", "3"], ["3", "3", "3"], ["3", "3", "3"]],
            [["2", "2"], ["2", "2"]],
            [["1"]],
        ],
        highlevel=False,
    )
    tags = ak.index.Index8(np.array([0, 1, 0, 1, 0, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2], dtype=np.int64))
    array = ak.contents.UnionArray(tags, index, [content1, content2])

    assert to_list(array) == [
        [1.1],
        [[3, 3, 3], [3, 3, 3], [3, 3, 3]],
        [2.2, 2.2],
        [[2, 2], [2, 2]],
        [3.3, 3.3, 3.3],
        [[1]],
    ]
    assert to_list(array[1:]) == [
        [[3, 3, 3], [3, 3, 3], [3, 3, 3]],
        [2.2, 2.2],
        [[2, 2], [2, 2]],
        [3.3, 3.3, 3.3],
        [[1]],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array)) == [
        1.1,
        [3, 3, 3],
        [3, 3, 3],
        [3, 3, 3],
        2.2,
        2.2,
        [2, 2],
        [2, 2],
        3.3,
        3.3,
        3.3,
        [1],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:])) == [
        [3, 3, 3],
        [3, 3, 3],
        [3, 3, 3],
        2.2,
        2.2,
        [2, 2],
        [2, 2],
        3.3,
        3.3,
        3.3,
        [1],
    ]

    array = ak.contents.UnionArray(tags, index, [content2, content3])

    assert to_list(array) == [
        [[3, 3, 3], [3, 3, 3], [3, 3, 3]],
        [["3", "3", "3"], ["3", "3", "3"], ["3", "3", "3"]],
        [[2, 2], [2, 2]],
        [["2", "2"], ["2", "2"]],
        [[1]],
        [["1"]],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [3, 3, 3, 3, 3, 3, 3, 3, 3],
        ["3", "3", "3", "3", "3", "3", "3", "3", "3"],
        [2, 2, 2, 2],
        ["2", "2", "2", "2"],
        [1],
        ["1"],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[1:], axis=2)) == [
        ["3", "3", "3", "3", "3", "3", "3", "3", "3"],
        [2, 2, 2, 2],
        ["2", "2", "2", "2"],
        [1],
        ["1"],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array[:, 1:], axis=2)) == [
        [3, 3, 3, 3, 3, 3],
        ["3", "3", "3", "3", "3", "3"],
        [2, 2],
        ["2", "2"],
        [],
        [],
    ]


def test_flatten_0042():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True
    ).layout
    assert ak.operations.to_list(ak.operations.flatten(array, axis=1)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]


def test_flatten2():
    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10], dtype=np.int64))
    array = ak.contents.ListOffsetArray(offsets, content)

    assert to_list(array) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(ak._do.flatten(array, axis=1)) == [
        0.0,
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]
    assert to_list(ak._do.flatten(array, axis=-1)) == [
        0.0,
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]
    with pytest.raises(ValueError) as err:
        assert to_list(ak._do.flatten(array, axis=-2))
        assert str(err.value).startswith("axis=0 not allowed for flatten")

    array2 = array[2:-1]
    assert to_list(ak._do.flatten(array2, axis=1)) == [3.3, 4.4, 5.5]
    assert to_list(ak._do.flatten(array2, axis=-1)) == [3.3, 4.4, 5.5]


def test_ByteMaskedArray_flatten():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)

    assert to_list(array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=1)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=-2)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        None,
        None,
        [10.0, 11.1, 12.2],
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=-1)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        None,
        None,
        [10.0, 11.1, 12.2],
    ]


def test_flatten_0198():
    array = ak.highlevel.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5]).layout

    assert ak.operations.to_list(ak.operations.flatten(array, axis=0)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]

    content0 = ak.operations.from_iter(
        [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5], highlevel=False
    )
    content1 = ak.operations.from_iter(
        ["one", None, "two", None, "three"], highlevel=False
    )
    array = ak.contents.UnionArray(
        ak.index.Index8(
            np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0], dtype=np.int8)
        ),
        ak.index.Index64(
            np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 5, 6, 4, 7], dtype=np.int64)
        ),
        [content0, content1],
    )

    assert to_list(array) == [
        1.1,
        "one",
        2.2,
        None,
        None,
        "two",
        3.3,
        None,
        None,
        None,
        4.4,
        "three",
        5.5,
    ]
    assert ak.operations.to_list(ak.operations.flatten(array, axis=0)) == [
        1.1,
        "one",
        2.2,
        "two",
        3.3,
        4.4,
        "three",
        5.5,
    ]


def test_fix_flatten_of_sliced_array_0446():
    array = ak.highlevel.Array([[1, 2, 3], [], [4, 5], [6, 7, 8, 9]]).layout

    assert ak.operations.flatten(array[:-1], axis=1).to_list() == [
        1,
        2,
        3,
        4,
        5,
    ]
    assert ak.operations.flatten(array[:-2], axis=1).to_list() == [1, 2, 3]
    assert ak.operations.flatten(array[:-1], axis=None).to_list() == [
        1,
        2,
        3,
        4,
        5,
    ]
    assert ak.operations.flatten(array[:-2], axis=None).to_list() == [
        1,
        2,
        3,
    ]


def test_flatten_None_axis():
    array2 = ak.highlevel.Array(np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7))

    assert to_list(ak.operations.flatten(array2, axis=None)) == to_list(
        np.arange(2 * 3 * 5 * 7)
    )


def test_fix_corner_case_0585():
    array = ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout

    assert ak.operations.flatten(array, axis=0).to_list() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]
    assert ak.operations.flatten(array, axis=-2).to_list() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]

    array = ak.highlevel.Array([1, 2, 3, 4, 5]).layout

    assert ak.operations.flatten(array, axis=0).to_list() == [
        1,
        2,
        3,
        4,
        5,
    ]
    assert ak.operations.flatten(array, axis=-1).to_list() == [
        1,
        2,
        3,
        4,
        5,
    ]


def test_flatten_allow_regulararray_size_zero_0590():
    empty = ak.contents.RegularArray(
        ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout, 0, zeros_length=0
    )

    assert ak.operations.flatten(empty, axis=0).to_list() == []
    assert ak.operations.flatten(empty, axis=1).to_list() == []
    assert ak.operations.flatten(empty, axis=2).to_list() == []


def test_0724():
    a = ak.contents.NumpyArray(np.empty(0))
    idx = ak.index.Index64([])
    a = ak.contents.IndexedOptionArray(idx, a)
    idx = ak.index.Index64([0])
    a = ak.contents.ListOffsetArray(idx, a)
    idx = ak.index.Index64([175990832])
    a = ak.contents.ListOffsetArray(idx, a)

    assert ak.operations.flatten(a, axis=2).to_list() == []
    assert str(ak.operations.flatten(a, axis=2).type) == "0 * var * ?float64"


def test_flatten_axis_none_0866():
    a1 = ak.operations.zip(
        {"a": [[1], [], [2, 3]], "b": [[4], [], [5, 6]]}, with_name="a1"
    )
    a2 = ak.operations.zip(
        {"a": [[7, 8], [9], []], "b": [[10, 11], [12], []]}, with_name="a2"
    )
    union = ak.operations.where([True, False, True], a1, a2)
    assert set(ak.operations.flatten(union, axis=None)) == {
        1,
        2,
        3,
        4,
        5,
        6,
        9,
        12,
    }


def test_0973():
    array = ak.operations.from_numpy(np.zeros((3, 3, 5)))
    flattened = ak.operations.flatten(array, axis=None)
    assert flattened.ndim == 1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_numpy_1d():
    x = ak.Array(
        ak.contents.NumpyArray(
            np.asarray([1, 2, 3]),
            parameters={"attrs": {"not": "hashable"}, "name": "x"},
        )
    )
    y = ak.Array(
        ak.contents.NumpyArray(
            np.asarray([1, 2, 3]),
            parameters={"attrs": {"not": "hashable"}, "name": "y"},
        )
    )
    result = x + y
    assert ak.parameters(result) == {"attrs": {"not": "hashable"}}


def test_numpy_2d():
    x = ak.Array(
        ak.contents.NumpyArray(
            np.asarray([[1, 2, 3]]),
            parameters={"attrs": {"not": "hashable"}, "name": "x"},
        )
    )
    y = ak.Array(
        ak.contents.NumpyArray(
            np.asarray([[1, 2, 3]]),
            parameters={"attrs": {"not": "hashable"}, "name": "y"},
        )
    )
    result = x + y
    assert ak.parameters(result) == {"attrs": {"not": "hashable"}}


def test_regular_array_numpy_1d():
    x = ak.Array(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(
                np.asarray([1, 2, 3, 4, 5, 6]),
                parameters={"attrs": {"not": "hashable"}, "name": "x"},
            ),
            3,
        )
    )
    y = ak.Array(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(
                np.asarray([7, 8, 9, 10, 11, 12]),
                parameters={"attrs": {"not": "hashable"}, "name": "y"},
            ),
            3,
        )
    )
    result = x + y
    assert result.layout.parameters == {}
    assert result.layout.content.parameters == {"attrs": {"not": "hashable"}}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
import numpy as np
import pytest

import awkward as ak
from awkward._behavior import behavior_of

to_list = ak.operations.to_list


def test():
    def _apply_ufunc(ufunc, method, inputs, kwargs):
        nextinputs = []
        for x in inputs:
            if (
                isinstance(x, ak.highlevel.Array)
                and x.layout.is_indexed
                and not x.layout.is_option
            ):
                nextinputs.append(
                    ak.highlevel.Array(x.layout.project(), behavior=behavior_of(x))
                )
            else:
                nextinputs.append(x)

        return getattr(ufunc, method)(*nextinputs, **kwargs)

    behavior = {}
    behavior[np.ufunc, "categorical"] = _apply_ufunc

    array = ak.highlevel.Array(
        ak.contents.IndexedArray(
            ak.index.Index64(np.array([0, 1, 2, 1, 3, 1, 4])),
            ak.contents.NumpyArray(np.array([321, 1.1, 123, 999, 2])),
            parameters={"__array__": "categorical"},
        ),
        behavior=behavior,
    )
    assert to_list(array * 10) == [3210, 11, 1230, 11, 9990, 11, 20]

    array = ak.highlevel.Array(["HAL"])
    with pytest.raises(TypeError):
        array + 1
